# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Main program file for the `lctime` standard-cell characterization tool.
"""

import os
import argparse
import joblib
import tempfile
from typing import List, Set

import liberty.parser as liberty_parser
from liberty.types import *
from ..cell_types import Combinational, SingleEdgeDFF, Latch, CellType

from PySpice.Unit import *

from ..logic.util import is_unate_in_xi
from ..liberty import util as liberty_util
from ..logic import functional_abstraction
from ..logic import seq_recognition
from ..logic.types import CombinationalOutput

from .util import *
from .timing_combinatorial import characterize_comb_cell
from .timing_sequential import *
from .input_capacitance import characterize_input_capacitances

from copy import deepcopy

from ..lccommon import net_util
from ..lccommon.net_util import load_transistor_netlist, is_ground_net, is_supply_net
import networkx as nx
from sympy.logic import satisfiable
import sympy.logic.boolalg

from PySpice.Spice.Parser import SpiceParser
import logging

from ..licence import licence_notice_string_single_line, original_source_repository


def _boolean_to_lambda(boolean: boolalg.Boolean):
    """
    Convert a sympy.boolalg.Boolean expression into a Python lambda function.
    :param boolean:
    :return:
    """
    simple = sympy.simplify(boolean)
    f = sympy.lambdify(boolean.atoms(), simple)
    return f


def recognize_cell_from_liberty(cell_group: Group) -> Union[Latch, SingleEdgeDFF, Combinational]:
    """
    Analyze the liberty group and derive the type of the cell.
    :param cell_group:
    :return:
    """

    # Get information on pins from the liberty file.
    input_pins, outputs_user = liberty_util.get_pin_information(cell_group)

    ff_group = cell_group.get_groups("ff")
    latch_group = cell_group.get_groups("latch")

    if ff_group and latch_group:
        logger.error("Cell contains a 'ff' and 'latch' description.")
        assert False, "Cannot characterize cells with both 'ff' and 'latch'."
    elif ff_group:
        if len(ff_group) != 1:
            assert False, "Cannot characterize cells with more than one 'ff' group."
        logger.info("'ff' group found. Cell is expected to be a flip-flop.")
        ff_group = ff_group[0]
        assert isinstance(ff_group, Group)

        # Get state names.
        iq, iqn = ff_group.args
        clocked_on = ff_group.get_boolean_function('clocked_on')
        next_state = ff_group.get_boolean_function('next_state')

        clear = ff_group.get_boolean_function('clear')
        preset = ff_group.get_boolean_function('preset')

        # TODO: Store and use this.
        clear_preset_var1 = ff_group['clear_preset_var1']
        clear_preset_var2 = ff_group['clear_preset_var2']

        # TODO: Store and use this.
        clocked_on_also = ff_group.get_boolean_function('clocked_on_also')

        cell_type = SingleEdgeDFF()
        cell_type.internal_state = sympy.Symbol(iq)
        cell_type.clocked_on = clocked_on
        cell_type.next_state = next_state
        cell_type.async_preset = preset
        cell_type.async_clear = clear

    elif latch_group:
        if len(latch_group) != 1:
            assert False, "Cannot characterize cells with more than one 'latch' group."
        logger.info("'latch' group found. Cell is expected to be a latch.")
        latch_group = latch_group[0]
        assert isinstance(latch_group, Group)

        # Get state names.
        iq, iqn = latch_group.args

        cell_type = Latch()
        cell_type.internal_state = sympy.Symbol(iq)

        clocked_on = latch_group.get_boolean_function('clocked_on')
        next_state = latch_group.get_boolean_function('next_state')

        clear = latch_group.get_boolean_function('clear')
        preset = latch_group.get_boolean_function('preset')

        # TODO: Store and use this.
        clear_preset_var1 = latch_group['clear_preset_var1']
        clear_preset_var2 = latch_group['clear_preset_var2']

        # TODO: Store and use this.
        clocked_on_also = latch_group.get_boolean_function('clocked_on_also')

        cell_type.clocked_on = clocked_on
        cell_type.next_state = next_state
        cell_type.async_preset = preset
        cell_type.async_clear = clear
    else:
        # No sequential element.
        cell_type = Combinational()

    # TODO:
    # power_down_function = cell_group.get_boolean_function('power_down_function')
    # cell_type.power_down_function = power_down_function

    # Copy description of output pins.
    for name, output in outputs_user.items():
        cell_type.outputs[sympy.Symbol(name)] = output

    input_pins = {sympy.Symbol(p) for p in input_pins}

    inputs_found_in_outputs = {i for output in cell_type.outputs.values()
                               for i in output.get_inputs()}
    diff = inputs_found_in_outputs - input_pins
    if diff:
        logger.warning(f"Some input pins are used but not declared in the liberty file: {sorted(diff)}")

    cell_type.inputs = sorted(input_pins, key=str)

    return cell_type


def abort(message: str, exit_code=1):
    """
    Exit the program due to an error.
    :param message: Error message.
    :param exit_code:
    """
    logger.error(message)
    exit(exit_code)


def main():
    """
    Command-line tool for cell characterization.
    Currently only combinatorial cells are supported excluding tri-state cells.
    :return:
    """

    print(licence_notice_string_single_line())
    print()

    logger = logging.getLogger(__name__)
    logger.info("lctime main function")

    parser = argparse.ArgumentParser(
        description='Characterize the timing of a combinatorial cell based on a SPICE netlist. '
                    'The resulting liberty file will contain the data of the input liberty file '
                    'plus the updated characteristics of the selected cell.',
        epilog='Example: lctime --liberty specification.lib --cell INVX1 AND2X1 --spice netlists.sp -I '
               'transistor_model.m --output mylib.lib')

    parser.add_argument('-l', '--liberty', required=True, metavar='LIBERTY', type=str,
                        help='Liberty file. This must contain all necessary specifications '
                             'needed to characterize the cell.')

    parser.add_argument('--cell', required=True, metavar='CELL_NAME', type=str,
                        action='append',
                        nargs='+',
                        help='Names of cells to be characterized.')

    parser.add_argument('--spice', required=True, metavar='SPICE', type=str,
                        action='append',
                        nargs='+',
                        help='SPICE netlist containing a subcircuit with the same name as the cell.')

    parser.add_argument('-I', '--include', required=False, action='append', metavar='SPICE_INCLUDE', type=str,
                        help='SPICE files to include such as transistor models.')

    parser.add_argument('-L', '--library', required=False, action='append', metavar='SPICE_LIB', type=str,
                        help='SPICE .LIB statements defining each a path to the library and a library name.'
                             'Example: --library "/path/to/lib libraryName".')

    parser.add_argument('--calc-mode', metavar='CALC_MODE', type=str, choices=['worst', 'typical', 'best'],
                        default='typical',
                        help='Calculation mode for computing the default timing arc'
                             ' based on the conditional timing arcs. "worst", "typical" (average) or "best".')

    parser.add_argument('-o', '--output', required=True, metavar='LIBERTY_OUT', type=str, help='Output liberty file.')

    parser.add_argument('--workingdir', required=False, metavar='WORKDIR', type=str,
                        help="Directory for ngspice simulation scripts and raw results.")

    parser.add_argument('--output-loads', required=True, metavar='CAPACITANCES', type=str,
                        help="List of output load capacitances for the cells. In pico Farads."
                             " List must be quoted, elements must be separated by a comma."
                             " Example: '0.05, 0.1, 0.2'")

    parser.add_argument('--slew-times', required=True, metavar='SLEWTIMES', type=str,
                        help="List of slew times of the input signals in nano seconds."
                             " List must be quoted, elements must be separated by a comma."
                             " Example: '0.05, 0.1, 0.2'")

    parser.add_argument('--related-pin-transition', required=False, metavar='SLEWTIMES', type=str,
                        help="List of slew times of the clock signal in nano seconds. "
                             "This is used for sequential cells only. "
                             "List must be quoted, elements must be separated by a comma. "
                             "Example: '0.05, 0.1, 0.2'")

    parser.add_argument('--analyze-cell-function', action='store_true',
                        help='Derive the logical function of the cell from the SPICE netlist (experimental).')

    parser.add_argument('--diff', required=False,
                        nargs="+",
                        metavar='DIFFERENTIAL_PATTERN',
                        type=str,
                        help='Specify differential inputs as "NonInverting,Inverting" tuples.'
                             'The placeholder "%%" can be used like "%%_P,%%_N" or "%%,%%_Diff", ...')

    parser.add_argument('--time-step', default=10e-12,
                        metavar='TIME_STEP',
                        type=float,
                        help='Specify the simulation time-step in seconds. Default is 10e-12.')

    parser.add_argument('--debug', action='store_true',
                        help='Enable debug mode (more verbose logging).')

    parser.add_argument('--debug-plots', action='store_true',
                        help='Create debug plots of simulation waveforms.')

    # Parse arguments
    args = parser.parse_args()

    DEBUG = args.debug
    log_level = logging.DEBUG if DEBUG else logging.INFO

    if DEBUG:
        log_format = '%(module)16s %(levelname)8s: %(message)s'
    else:
        # Also output name of function in DEBUG mode.
        log_format = '%(module)16s %(funcName)16s %(levelname)8s: %(message)s'

    logging.basicConfig(format=log_format, level=log_level)

    workingdir = args.workingdir
    if workingdir is None:
        workingdir = tempfile.mkdtemp(prefix="lctime-")

    # Get list of cell names to be characterized.
    cell_names = [n for names in args.cell for n in names]  # Flatten the nested list.

    # Get list of user-provided netlist files.
    netlist_files = [n for names in args.spice for n in names]  # Flatten the nested list.

    # Generate a lookup-table which tells for each cell name which netlist file to use.
    netlist_file_table: Dict[str, str] = dict()
    for netlist_file in netlist_files:
        logger.info("Load SPICE netlist: {}".format(netlist_file))
        parser = SpiceParser(path=netlist_file)
        for sub in parser.subcircuits:
            if sub.name in netlist_file_table:
                # Abort if a sub circuit is defined in multiple netlists.
                abort(
                    f"Sub-circuit '{sub.name}' is defined in multiple netlists: {netlist_file_table[sub.name]}, {netlist_file}")
            netlist_file_table[sub.name] = netlist_file

    # Test if all cell names can be found in the netlist files.
    cell_names_not_found = set(cell_names) - netlist_file_table.keys()
    if cell_names_not_found:
        abort(f"Cell names not found in netlists: {', '.join(cell_names_not_found)}")

    # Load liberty file.
    lib_file = args.liberty
    logger.info("Reading liberty: {}".format(lib_file))
    with open(lib_file) as f:
        data = f.read()
    library = liberty_parser.parse_liberty(data)

    # Check if the delay model is supported.
    delay_model = library['delay_model']
    supported_delay_models = ['table_lookup']
    if delay_model not in supported_delay_models:
        msg = "Delay model not supported: '{}'. Must be one of {}.".format(delay_model,
                                                                           ", ".join(supported_delay_models))
        logger.error(msg)
        assert False, msg

    # Make independent copies of the library.
    new_library = deepcopy(library)
    # Strip all cell groups.
    new_library.groups = [g for g in new_library.groups if g.group_name != 'cell']
    # Strip away all LUT template table groups.
    table_types = ['lu_table_template', 'power_lut_template']
    new_library.groups = [g for g in new_library.groups if g.group_name not in table_types]

    # Load operation voltage and temperature.
    # TODO: load voltage/temperature from operating_conditions group
    supply_voltage = library['nom_voltage']
    temperature = library['nom_temperature']
    logger.info('Supply voltage = {:f} V'.format(supply_voltage))
    logger.info('Temperature = {:f} V'.format(temperature))

    # Get timing corner from liberty file.
    # Find definitions of operating conditions and sort them by name.
    operating_conditions_list = library.get_groups('operating_conditions')
    # Put into a dict by name.
    operating_conditions: Dict[str, Group] = {str(g.args[0]): g for g in operating_conditions_list}

    logger.info("Operating conditions: {}".format(set(operating_conditions.keys())))

    """
    TODO: Use the information from the operating conditions.
    Example:
    operating_conditions (MPSS) {
        calc_mode : worst ;
        process : 1.5 ;
        process_label : "ss" ;
        temperature : 70 ;
        voltage : 4.75 ;
        tree_type : worse_case_tree ;
    }
    """

    def _transistors2multigraph(transistors) -> nx.MultiGraph:
        """ Create a graph representing the transistor network.
            Each edge corresponds to a transistor, each node to a net.
        """
        G = nx.MultiGraph()
        for t in transistors:
            G.add_edge(t.source_net, t.drain_net, (t.gate_net, t.channel_type))
        assert nx.is_connected(G)
        return G

    # Get timing corner from liberty file.
    # TODO: let user overwrite it.
    calc_modes = {
        'typical': CalcMode.TYPICAL,
        'worst': CalcMode.WORST,
        'best': CalcMode.BEST,
    }

    # TODO: Make use of this.
    default_operating_conditions = library['default_operating_conditions']
    logger.info("Default operating conditions: {}".format(default_operating_conditions))

    assert args.calc_mode in calc_modes, "Unknown calculation mode: {}".format(args.calc_mode)

    calc_mode = calc_modes[args.calc_mode]
    logger.info("calc_mode: {}".format(calc_mode.name))

    # Read trip points from liberty file.
    trip_points = read_trip_points_from_liberty(library)

    logger.debug(trip_points)

    spice_includes = args.include if args.include else []
    if len(spice_includes) == 0:
        logger.warning("No transistor model supplied. Use --include or -I.")

    # Sanitize include paths.
    input_argument_error = False
    for path in spice_includes:
        if not os.path.isfile(path):
            logger.error(f"Include file does not exist: {path}")
            input_argument_error = True

    spice_libraries_raw: List[str] = args.library if args.library else []
    # Split library statements into path and library name.
    spice_libraries: List[Tuple[str, str]] = [tuple(s.strip() for s in l.split(" ", maxsplit=2))
                                              for l in spice_libraries_raw
                                              ]
    # Sanitize the library arguments.
    for lib, raw in zip(spice_libraries, spice_libraries_raw):
        if len(lib) != 2 or not lib[0] or not lib[1]:
            abort(f'Library statements must be of the format "/path/to/library libraryName". Found: "{raw}".')

        path, name = lib
        if not os.path.isfile(path):
            logger.error(f"Library file does not exist: {path}")
            input_argument_error = True

    # Exit if some input arguments were obviously invalid.
    if input_argument_error:
        abort("Exit because of invalid arguments.")

    # .LIB statements
    library_statements = [f".LIB {path} {name}" for path, name in spice_libraries]

    # .INCLUDE statements
    include_statements = [f".include {i}" for i in spice_includes]

    setup_statements = library_statements + include_statements

    # Setup array of output capacitances and input slews.
    output_capacitances = np.array([float(s.strip()) for s in args.output_loads.split(",")]) * 1e-12  # pF
    input_transition_times = np.array([float(s.strip()) for s in args.slew_times.split(",")]) * 1e-9  # ns

    # Transition times of the clock pin.
    if args.related_pin_transition:
        related_pin_transition = np.array(
            [float(s.strip()) for s in args.related_pin_transition.split(",")]) * 1e-9  # ns
    else:
        related_pin_transition = None

    logger.info(f"Output capacitances [pF]: {output_capacitances / 1e-12}")
    logger.info(f"Input slew times [ns]: {input_transition_times / 1e-9}")
    if related_pin_transition is not None:
        logger.info(f"Related pin transition times [ns]: {related_pin_transition / 1e-9}")

    # TODO: Make time resolution parametrizable.
    time_resolution_seconds = float(args.time_step)
    logger.info(f"Time resolution = {time_resolution_seconds}s")
    if time_resolution_seconds <= 0:
        abort('Time step must be larger than zero.')

    if time_resolution_seconds > 1e-9:
        logger.warning(f"Timestep is larger than 1ns: {time_resolution_seconds}s")

    # Setup configuration struct.
    conf = CharacterizationConfig()
    conf.supply_voltage = supply_voltage
    conf.trip_points = trip_points
    conf.timing_corner = calc_mode
    conf.setup_statements = setup_statements
    conf.time_step = time_resolution_seconds
    conf.temperature = temperature
    conf.workingdir = workingdir
    conf.debug = args.debug
    conf.debug_plots = args.debug_plots

    conf.get_units_from_liberty(library)

    logger.info(f"Capacitance unit: {conf.capacitance_unit} F")
    logger.info(f"Time unit: {conf.time_unit} s")

    # Characterize all cells in the list.
    def characterize_cell(cell_name: str) -> Group:
        """
        Characterize a cell and create an updated cell group.
        :param cell_name:
        :return: Return an updated cell group.
        """

        # Create working directory if it does not exist yet.
        cell_workingdir = os.path.join(conf.workingdir, cell_name)
        if not os.path.exists(cell_workingdir):
            os.mkdir(cell_workingdir)

        # Get netlist and liberty group.
        netlist_file = netlist_file_table[cell_name]
        try:
            cell_group = select_cell(library, cell_name)

            # Determine type of cell (latch, flip-flop, combinational).
            cell_type_liberty = recognize_cell_from_liberty(cell_group)
            cell_type = cell_type_liberty
        except KeyError as e:
            logger.warning(f"No cell group defined yet in liberty file: {cell_name}")

            if not args.analyze_cell_function:
                abort("Cell is not defined in liberty. Enable cell recognition with --analyze.")

            cell_type_liberty = Combinational()  # Default: Empty cell.

            # Cell group does not exist, so create it.
            logger.debug("Create empty cell group.")
            cell_group = Group(group_name='cell', args=[cell_name])
            library.groups.append(cell_group)

            # The liberty did not define anything about this cell.
            cell_type = None

        # Check that the name matches.
        assert cell_group.args == [cell_name], "Cell name does not match."  # This should not happen.

        logger.info("Cell: {}".format(cell_name))
        logger.info("Netlist: {}".format(netlist_file))

        # Get information on pins from the liberty file.
        input_pins = [str(s) for s in cell_type_liberty.inputs]
        output_pins = [str(s) for s in cell_type_liberty.outputs.keys()]

        liberty_pins = set(input_pins) | set(output_pins)
        # Convert to strings.
        liberty_pins = {str(p) for p in liberty_pins}

        # Create a lookup table to reconstruct lower/upper case letters.
        # This is a workaround. The SPICE parser converts everything to uppercase.
        case_lookup_table = {p.lower(): p for p in liberty_pins}
        if len(case_lookup_table) != len(liberty_pins):
            # It's not a one-to-one mapping!
            logger.warning(f"Mixed lower case and upper case could cause trouble.")

        def fix_case(pin: str) -> str:
            """
            Restore lower/upper case of signals that went lost during SPICE parsing.
            """
            return case_lookup_table.get(pin.lower(), pin)

        logger.info(f"Input pins as defined in liberty: {input_pins}")
        logger.info(f"Output pins as defined in liberty: {sorted(cell_type_liberty.outputs.keys())}")

        # Load netlist of cell
        # TODO: Load all netlists at the beginning.
        logger.info('Load netlist: %s', netlist_file)
        try:
            transistors_abstract, cell_pins = load_transistor_netlist(netlist_file, cell_name)
        except Exception as e:
            abort(str(e))

        if len(transistors_abstract) == 0:
            msg = "No transistors found in cell. (The netlist must be flattened, sub-circuits are not resolved)"
            abort(msg)

        cell_pins = [fix_case(p) for p in cell_pins]
        for t in transistors_abstract:
            t.source_net = fix_case(t.source_net)
            t.drain_net = fix_case(t.drain_net)
            t.gate_net = fix_case(t.gate_net)

        # Get pin ordering of spice circuit.
        spice_ports = get_subcircuit_ports(netlist_file, cell_name)
        logger.info(f"SPICE subcircuit ports: {spice_ports}")
        io_pins = net_util.get_io_pins(cell_pins)

        # Detect power pins.
        # TODO: don't decide based only on net name.
        power_pins = [p for p in cell_pins if net_util.is_power_net(p)]
        assert len(power_pins) == 2, "Expected to have 2 power pins."
        vdd_pins = [p for p in power_pins if net_util.is_supply_net(p)]
        gnd_pins = [p for p in power_pins if net_util.is_ground_net(p)]
        assert len(vdd_pins) == 1, "Expected to find one VDD pin but found: {}".format(vdd_pins)
        assert len(gnd_pins) == 1, "Expected to find one GND pin but found: {}".format(gnd_pins)
        vdd_pin = vdd_pins[0]
        gnd_pin = gnd_pins[0]

        # Sanity check: All pins defined in liberty must appear in the SPICE netlist.
        all_liberty_pins = set()
        for pin in cell_group.get_groups("pin"):
            assert isinstance(pin, liberty_parser.Group)
            pin_name = pin.args[0]
            all_liberty_pins.add(pin_name)
            complementary_pin = pin.get("complementary_pin")
            if complementary_pin is not None:
                all_liberty_pins.add(complementary_pin)
        all_spice_pins = set(cell_pins)
        pins_not_in_spice = sorted(all_liberty_pins - all_spice_pins)

        if pins_not_in_spice:
            abort(f"Pins defined in liberty but not in SPICE netlist: {', '.join(pins_not_in_spice)}")

        # Convert the transistor network into its multi-graph representation.
        # This is used for a formal analysis of the network.
        transistor_graph = _transistors2multigraph(transistors_abstract)

        # Detect input nets from the transistor netlist (if enabled).
        if args.analyze_cell_function:
            logger.debug("Detect input nets from the circuit.")
            detected_inputs = functional_abstraction.find_input_gates(transistor_graph)
            # Detect nets that must be inputs (connected to gates only but do not
            # appear in the list of pins in the SPICE circuit definition.
            inputs_missing_in_spice = detected_inputs - all_spice_pins
            if inputs_missing_in_spice:
                logger.warning(f"The circuit has gate nets that must be inputs "
                               f"but are not in the pin definition of the SPICE circuit: "
                               f"{', '.join(sorted(inputs_missing_in_spice))}")
            # Same check for pins declared in liberty template.
            inputs_missing_in_liberty = detected_inputs - all_liberty_pins
            if inputs_missing_in_liberty:
                logger.warning(f"The circuit has gate nets that must be inputs "
                               f"but are not declared as a pin in the liberty template: "
                               f"{', '.join(sorted(inputs_missing_in_liberty))}")

            # Add detected input pins.
            diff = detected_inputs - set(input_pins)
            if diff:
                logger.info(f"Also include detected pins: {', '.join(sorted(diff))}")
                input_pins.extend(diff)

            # Find pins that are defined in the SPICE circuit but are not inputs nor power.
            maybe_outputs = all_spice_pins - set(input_pins) - set(power_pins)
            if maybe_outputs:
                logger.info(f"Potential output pins: {', '.join(sorted(maybe_outputs))}")
                output_pins.extend(maybe_outputs)

        # Sanity check.
        if len(input_pins) == 0:
            msg = "Cell has no input pins."
            logger.error(msg)
            assert False, msg

        # Sanity check.
        if len(output_pins) == 0:
            msg = "Cell has no output pins."
            logger.error(msg)
            assert False, msg

        # Extract differential pairs from liberty.
        # Liberty allows to reference complementary pins with the 'complementary_pin' attribute.
        logger.debug("Load complementary pins from liberty.")
        differential_inputs_liberty = dict()
        for pin in cell_group.get_groups("pin"):
            assert isinstance(pin, liberty_parser.Group)
            pin_name = pin.args[0]
            complementary_pin = pin.get("complementary_pin")
            if complementary_pin is not None:
                differential_inputs_liberty[pin_name] = complementary_pin

        # Match differential inputs.
        if args.diff is not None:
            logger.debug("Match complementary pins from user-defined pattern.")
            differential_inputs_from_pattern = find_differential_inputs_by_pattern(args.diff, input_pins)
        else:
            differential_inputs_from_pattern = dict()

        # Merge the differential pairs provided from the user with the pairs detected from the liberty file.
        differential_inputs_liberty.update(differential_inputs_from_pattern)
        differential_inputs = differential_inputs_liberty

        # Sanity checks on complementary pins.
        # Complementary pin should not be defined as pin group in liberty file.
        for pin in cell_group.get_groups("pin"):
            assert isinstance(pin, liberty_parser.Group)
            pin_name = pin.args[0]
            if pin_name in differential_inputs.values():
                logger.warning(
                    f"Complementary pin is modelled in the liberty file but will not be characterized: {pin_name}")

        for noninv, inv in differential_inputs.items():
            logger.info(f"Differential input (+,-): {noninv}, {inv}")

        # Find all input pins that are not inverted inputs of a differential pair.
        inverted_pins = differential_inputs.values()
        input_pins_non_inverted = [p for p in input_pins if p not in inverted_pins]
        "All input pins that are not inverted inputs of a differential pair"

        if args.analyze_cell_function:
            # Derive boolean functions for the outputs from the netlist.
            logger.info("Derive boolean functions for the outputs based on the netlist.")

            cell_type = analyze_boolean_functions(
                cell_type,
                cell_group,
                transistor_graph,
                io_pins,
                differential_inputs,
                vdd_pin,
                gnd_pin,
            )

            merge_cell_types(cell_type_liberty, cell_type)
        else:
            # Skip functional abstraction and take the functions provided in the liberty file.
            # output_functions_symbolic = output_functions_user
            pass  # TODO

        logger.info("Run characterization.")

        # Setup cell specific configuration.
        cell_conf = CellConfig()
        cell_conf.cell_name = cell_name
        cell_conf.global_conf = conf
        cell_conf.complementary_pins = differential_inputs
        cell_conf.ground_net = gnd_pin
        cell_conf.supply_net = vdd_pin
        cell_conf.workingdir = cell_workingdir
        cell_conf.spice_netlist_file = netlist_file_table[cell_name]
        cell_conf.spice_ports = spice_ports

        # Add groups for the cell to be characterized.
        new_cell_group = deepcopy(select_cell(library, cell_name))

        # Populate the cell group with missing pin groups.
        create_missing_pin_groups(
            new_cell_group,
            cell_type,
            cell_conf,
            output_pins,
            input_pins_non_inverted
        )

        # Measure input pin capacitances.
        measure_input_capacitances(
            conf,
            cell_conf,
            input_pins,
            input_pins_non_inverted,
            output_pins,
            new_cell_group
        )

        if isinstance(cell_type, Combinational):
            characterize_combinational_output(
                new_library,
                new_cell_group,
                conf,
                cell_type,
                cell_conf,
                input_pins,
                input_pins_non_inverted,
                output_capacitances,
                input_transition_times
            )
        elif isinstance(cell_type, SingleEdgeDFF):
            characterize_flip_flop_output(
                new_library,
                new_cell_group,
                conf,
                cell_type,
                cell_conf,
                related_pin_transition,
                input_transition_times,
                output_capacitances=output_capacitances
            )
        elif isinstance(cell_type, Latch):
            logger.info("Characterize latch.")
            logger.error("Characterization of latches is not yet supported.")

            """
            Characterization of latches is very similar to the one of flip-flops. Delays, hold and setup times
            are measured relative to the de-activating edge of the clock signal instead of the active clock edge.
            """

            assert False, "Characterization of latches is not yet supported."
        else:
            assert False, f"Unsupported cell type: {type(cell_type)}"

        assert isinstance(new_cell_group, Group)
        return new_cell_group

    # Characterize cells in parallel.
    # new_cell_groups = joblib.Parallel(n_jobs=-1, prefer='threads') \
    #     (joblib.delayed(characterize_cell)(cell_name) for cell_name in cell_names)

    # Characterize cells sequentially.
    new_cell_groups = [characterize_cell(cell_name) for cell_name in cell_names]

    for new_cell_group in new_cell_groups:
        new_library.groups.append(new_cell_group)

    with open(args.output, 'w') as f:
        logger.info("Write liberty: {}".format(args.output))

        # Write URL to lctime.

        s = """/*
Generated by lctime ({}).
*/

"""
        f.write(s.format(original_source_repository()))

        f.write(str(new_library))


def analyze_boolean_functions(
        cell_type: CellType,
        cell_group: Group,
        transistor_graph: nx.MultiGraph,
        io_pins: Set,
        differential_inputs: Dict[str, str],
        vdd_pin: str,
        gnd_pin: str,
) -> CellType:
    abstracted_circuit = functional_abstraction.analyze_circuit_graph(
        graph=transistor_graph,
        pins_of_interest=io_pins,
        constant_input_pins={
            vdd_pin: True,
            gnd_pin: False},
        differential_inputs=differential_inputs,
        user_input_nets=None
    )

    if abstracted_circuit.latches:
        # There's some feedback loops in the circuit.

        # Try to recognize sequential cells.
        detected_cell_type = seq_recognition.extract_sequential_circuit(abstracted_circuit)

        if detected_cell_type:
            logger.info(f"Detected sequential circuit:\n{detected_cell_type}")

    else:
        logger.info("Detected purely combinational circuit.")
        detected_cell_type = Combinational()
        detected_cell_type.outputs = abstracted_circuit.outputs
        detected_cell_type.inputs = abstracted_circuit.get_primary_inputs()
        if cell_type is None:
            cell_type = Combinational()

    if cell_type is None or len(cell_group.groups) == 0:
        cell_type = detected_cell_type
    else:
        # Sanity check: Detected cell type (combinational, latch, ff) must match with the liberty file.
        if type(detected_cell_type) is not type(cell_type):
            msg = f"Mismatch: Detected cell type is {type(detected_cell_type)} " \
                  f"but liberty says {type(cell_type)}."
            logger.error(msg)
            assert False, msg

    output_functions_deduced = abstracted_circuit.outputs

    # Log deduced output functions.
    for output_name, value in output_functions_deduced.items():
        if value.high_impedance:
            logger.info(
                f"Deduced output function: {output_name} = {value.function}, tri_state = {value.high_impedance}")
        else:
            logger.info(f"Deduced output function: {output_name} = {value.function}")

    # # Convert keys into strings (they are `sympy.Symbol`s now)
    # output_functions_deduced = {str(output.name): comb.function for output, comb in
    #                             output_functions_deduced.items()}
    # output_functions_symbolic = output_functions_deduced.copy()

    return cell_type


def merge_cell_types(
        source_cell_type: CellType,
        target_cell_type: CellType
):
    # Merge deduced output functions with the ones read from the liberty file and perform consistency check.
    for output_symbol, output in source_cell_type.outputs.items():
        output_name = str(output_symbol)
        logger.info(f"User supplied output function: {output_name} = {output_name}")

        assert output_symbol in target_cell_type.outputs, f"No function has been deduced for output pin '{output_name}'."
        # Consistency check:
        # Verify that the deduced output formula is equal to the one defined in the liberty file.
        logger.info("Check equality of boolean function in liberty file and derived function.")
        equal = functional_abstraction.bool_equals(output.function,
                                                   target_cell_type.outputs[output_symbol].function)
        if not equal:
            msg = "User supplied function does not match the deduced function for pin '{}'".format(output_name)
            logger.error(msg)

        if equal:
            # Take the function defined by the liberty file.
            # This might be desired because it is in another form (CND, DNF,...).
            target_cell_type.outputs[output_symbol] = output


def create_missing_pin_groups(
        new_cell_group: Group,
        cell_type: CellType,
        cell_conf: CellConfig,
        output_pins: List[str],
        input_pins_non_inverted: List[str]
):
    # Strip away timing groups. They will be replaced by the new characterization.
    for pin_group in new_cell_group.get_groups('pin'):
        pin_group.groups = [g for g in pin_group.groups if g.group_name != 'timing']

    # Create missing pin groups.
    for pin in sorted(set(input_pins_non_inverted + output_pins)):
        pin_group = new_cell_group.get_groups('pin', pin)
        if not pin_group:
            pin_group = Group('pin', args=[pin])
            new_cell_group.groups.append(pin_group)

    # Set 'direction' attribute of input pins.
    for pin in input_pins_non_inverted:
        pin_group = new_cell_group.get_group('pin', pin)
        if 'direction' not in pin_group:
            pin_group['direction'] = 'input'

    # Set 'direction' attribute of output pins.
    for pin in output_pins:
        pin_group = new_cell_group.get_group('pin', pin)
        if 'direction' not in pin_group:
            pin_group['direction'] = 'output'
        pin_symbol = sympy.Symbol(pin)
        if cell_type.outputs[pin_symbol].is_tristate():
            # Mark as tri-state.
            pin_group.set_boolean_function('three_state', cell_type.outputs[pin_symbol].high_impedance)

    # Create 'complementary_pin' attribute for the inverted pin of differential pairs.
    for input_pin in input_pins_non_inverted:
        input_pin_group = new_cell_group.get_group('pin', input_pin)
        # Create link to inverted pin for differential inputs.
        input_pin_inverted = cell_conf.complementary_pins.get(input_pin)
        if input_pin_inverted:
            input_pin_group['complementary_pin'] = [EscapedString(input_pin_inverted)]


def measure_input_capacitances(
        conf: CharacterizationConfig,
        cell_conf: CellConfig,
        input_pins: List[str],
        input_pins_non_inverted: List[str],
        output_pins: List[str],
        cell_group: Group
):
    """
    Measure input capacitances and write them into the `cell_group`.
    :param conf: 
    :param cell_conf: 
    :param input_pins: 
    :param input_pins_non_inverted: 
    :param output_pins: 
    :param cell_group: 
    :return: 
    """
    logger.debug(f"Measuring input pin capacitances of cell {cell_conf.cell_name}.")
    for input_pin in input_pins_non_inverted:
        # Input capacitances are not measured for the inverting inputs of differential pairs.
        logger.info("Measuring input capacitance: {} {}".format(cell_conf.cell_name, input_pin))
        input_pin_group = cell_group.get_group('pin', input_pin)

        result = characterize_input_capacitances(
            input_pins=input_pins,
            active_pin=input_pin,
            output_pins=output_pins,
            cell_conf=cell_conf
        )

        input_pin_group['rise_capacitance'] = result['rise_capacitance'] / conf.capacitance_unit
        input_pin_group['fall_capacitance'] = result['fall_capacitance'] / conf.capacitance_unit
        input_pin_group['capacitance'] = result['capacitance'] / conf.capacitance_unit


def characterize_combinational_output(
        new_library: Group,
        new_cell_group: Group,
        conf: CharacterizationConfig,
        cell_type: CellType,
        cell_conf: CellConfig,
        input_pins: List[str],
        input_pins_non_inverted: List[str],
        output_capacitances: np.ndarray,
        input_transition_times: np.ndarray
):
    assert isinstance(cell_type, Combinational) or isinstance(cell_type, SingleEdgeDFF)
    
    is_clock_to_output_arc = isinstance(cell_type, SingleEdgeDFF)
    
    # Measure timing for all input-output arcs.
    logger.debug("Measuring delay arcs.")
    for output_pin_symbol in cell_type.outputs.keys():
        output_pin_name = str(output_pin_symbol)
        output_pin_group = new_cell_group.get_group('pin', output_pin_name)

        # Store boolean function of output to liberty.
        output_pin_group.set_boolean_function('function', cell_type.outputs[output_pin_symbol].function)

        # Check if the output can be high-impedance.
        is_tristate = cell_type.outputs[output_pin_symbol].is_tristate()

        # Store three_state function to liberty.
        constant_input_pins = dict()
        if is_tristate:
            output_pin_group.set_boolean_function(
                'three_state', cell_type.outputs[output_pin_symbol].high_impedance
            )

            # Find normal operating conditions such that the output is not tri-state.
            models = list(satisfiable(~cell_type.outputs[output_pin_symbol].high_impedance, all_models=True))
            for model in models:
                logger.info(f"Output '{output_pin_name}' is low-impedance when {model}.")

            if len(models) > 1:
                abort("Characterization of tri-state outputs is not supported when the tri-state depends on"
                      "more than one input.")

            model = models[0]
            for name, value in model.items():
                constant_input_pins[str(name)] = value

        # Skip measuring the tri-state enable input.
        related_pins = sorted(set(input_pins_non_inverted) - constant_input_pins.keys())
        _input_pins = sorted(set(input_pins) - constant_input_pins.keys())

        # Characterize each from related_pin to output_pin.
        for related_pin in related_pins:

            related_pin_inverted = cell_conf.complementary_pins.get(related_pin)
            if related_pin_inverted:
                logger.info("Timing arc (differential input): ({}, {}) -> {}"
                            .format(related_pin, related_pin_inverted, output_pin_name))
            else:
                logger.info("Timing arc: {} -> {}".format(related_pin, output_pin_name))

            # Convert deduced output functions into Python lambda functions.
            output_functions = {
                str(name): _boolean_to_lambda(comb_output.function)
                for name, comb_output in cell_type.outputs.items()
            }

            # Get timing sense of this arc.
            timing_sense = str(is_unate_in_xi(output_functions[output_pin_name], related_pin).name).lower()
            logger.info("Timing sense: {}".format(timing_sense))

            result = characterize_comb_cell(
                input_pins=_input_pins,
                output_pin=output_pin_name,
                related_pin=related_pin,
                output_functions=output_functions,

                total_output_net_capacitance=output_capacitances,
                input_net_transition=input_transition_times,

                cell_conf=cell_conf,
                constant_inputs=constant_input_pins
            )

            # Get the table indices.
            # TODO: get correct index/variable mapping from liberty file.
            index_1 = result['total_output_net_capacitance'] / conf.capacitance_unit
            index_2 = result['input_net_transition'] / conf.time_unit

            # Create template tables.
            template_table = liberty_util.create_delay_template_table(new_library, len(index_1), len(index_2))
            table_template_name = template_table.args[0]

            # Create liberty timing tables.
            timing_tables = []
            for table_name in ['cell_rise', 'cell_fall', 'rise_transition', 'fall_transition']:
                table = Group(
                    table_name,
                    args=[table_template_name],
                )

                table.set_array('index_1', index_1)
                table.set_array('index_2', index_2)
                table.set_array('values', result[table_name] / conf.time_unit)

                timing_tables.append(table)

            # Create the liberty timing group.
            timing_attributes = [
                Attribute('related_pin', EscapedString(related_pin)),
                Attribute('timing_sense', timing_sense)
            ]

            timing_group = Group(
                'timing',
                attributes=timing_attributes,
                groups=timing_tables
            )

            # Attach timing group to output pin group.
            output_pin_group.groups.append(timing_group)


def characterize_flip_flop_output(
        new_library: Group,
        new_cell_group: Group,
        conf: CharacterizationConfig,
        cell_type: SingleEdgeDFF,
        cell_conf: CellConfig,
        related_pin_transition: np.ndarray,
        input_transition_times: np.ndarray,
        output_capacitances: np.ndarray,
):
    assert isinstance(cell_type, SingleEdgeDFF)

    logger.info("Characterize single-edge triggered flip-flop.")

    if related_pin_transition is None:
        abort("Need to specify 'related-pin-transition' for the clock pin.")

    # Create or update the 'ff' group.
    ff_group = new_cell_group.get_groups('ff')
    if not ff_group:
        ff_group = Group('ff')
        new_cell_group.groups.append(ff_group)

    # Store content of 'ff' group.
    ff_group.args = [str(cell_type.internal_state), str(cell_type.internal_state) + "_INV"]
    ff_group.set_boolean_function('clocked_on', cell_type.clocked_on)
    ff_group.set_boolean_function('next_state', cell_type.next_state)
    if cell_type.async_preset:
        ff_group.set_boolean_function('preset', cell_type.async_preset)
    if cell_type.async_clear:
        ff_group.set_boolean_function('clear', cell_type.async_clear)

    # Find clock pin.
    clock_signals = list(cell_type.clocked_on.atoms(sympy.Symbol))
    if len(clock_signals) != 1:
        logger.error(f"Expect exactly one clock signal. Got {clock_signals}")
    clock_signal = clock_signals[0]
    # Find clock polarity:
    clock_edge_polarity = cell_type.clocked_on.subs({clock_signal: True})
    clock_pin = str(clock_signal.name)

    assert isinstance(clock_pin, str)

    logger.info(f"Clock signal: {clock_pin}")
    logger.info(f"Clock polarity: {'rising' if clock_edge_polarity else 'falling'}")

    # Find preset/clear signals.
    # Make sure preset/clear are disabled.
    preset_condition = cell_type.async_preset
    clear_condition = cell_type.async_clear
    # Find a variable assignment such that neither preset nor clear is active.
    no_preset_no_clear = list(satisfiable(~preset_condition & ~clear_condition, all_models=True))
    for model in no_preset_no_clear:
        logger.info(f"FF in normal operation mode when: {model}")
    preset_clear_input = no_preset_no_clear[0]
    if len(no_preset_no_clear) > 1:
        logger.warning(f"Multiple possibilities found for disabling preset and clear. "
                       f"Take the first one ({preset_clear_input}).")

    # Find all data pins that are relevant for the internal state of the flip-flop.
    data_in_pins = sorted(cell_type.next_state.atoms(sympy.Symbol))
    logger.debug(f"Input pins relevant for internal state: {data_in_pins}")

    assert isinstance(cell_type.internal_state,
                      sympy.Symbol), "Internal flip-flop-state variable is not defined."

    # Find all output pins that depend on the internal state.
    data_out_pins: List[sympy.Symbol] = [name for name, output in cell_type.outputs.items()
                                         if cell_type.internal_state in output.function.atoms()
                                         ]
    logger.debug(f"Output pins that depend on the internal state: {data_out_pins}")

    
    # Remember a clock pulse width which for sure samples a stable data signal.
    clock_pulse_width_guess = 0

    # TODO use this function to simplify the below code.
    ## Find data input pins which can control the output pin.
    #control_and_observe_pins = find_flipflop_state_control_and_observe_pins(cell_type, preset_clear_input)
    #assert len(control_and_observe_pins) > 0, "No input pin found which can control the flip-flop state."
    
    
    # Characterize setup/hold for each data pin.
    for i, data_in_pin in enumerate(data_in_pins):
        logger.info(f"Measure constraints of pin {data_in_pin} ({i}/{len(data_in_pins)}).")

        # Find all assignments of the other data pins such that the data pin controls
        # the internal state.
        # Find values of the other pins such that:
        #  next_state(data_in_pin=0, other_pins) != next_state(data_in_pin=1, other_pins)
        next_state_0 = cell_type.next_state.subs({data_in_pin: False})
        next_state_1 = cell_type.next_state.subs({data_in_pin: True})
        models = list(satisfiable(next_state_0 ^ next_state_1, all_models=True))

        for other_pin_values in models:
            # Express the assignment of the other pins as a boolean formula.
            # This will also be used as a 'when' statement in the liberty file.
            when_other_inputs = sympy.And(*(pin if value else ~pin for pin, value in other_pin_values.items()))
            logger.info(f"Measure constraints of pin {data_in_pin} when {when_other_inputs}.")

            # Set static voltages of other input pins.
            other_pin_values.update(preset_clear_input)
            static_input_voltages = dict()
            for pin, value in other_pin_values.items():
                if not isinstance(pin, sympy.Symbol):
                    continue
                value = value == True
                voltage = cell_conf.global_conf.supply_voltage if value else 0.0
                pin = str(pin)
                logger.debug(f"{pin} = {voltage} V")
                static_input_voltages[pin] = voltage

            # Find an output pin such that the internal state is observable.
            observer_outputs = []  # Output pins that can observe the internal memory state.
            for output_pin, output in cell_type.outputs.items():
                function = output.function
                # Substitute with constant input pins.
                function = function.subs(other_pin_values)

                # Compute the output for all values of the internal state and make sure it is different.
                function0 = function.subs({cell_type.internal_state: False})
                function1 = function.subs({cell_type.internal_state: True})
                is_observable = not satisfiable(~(function0 ^ function1))  # Test if function0 != function1

                logger.debug(f"Internal state {cell_type.internal_state} observable from output {output_pin} "
                             f"when {other_pin_values}: {is_observable}")

                if is_observable:
                    observer_outputs.append(output_pin)

            logger.debug(f"Internal state is observable from: {observer_outputs}")

            if not observer_outputs:
                # When the internal state is not observable we cannot measure constraints.
                # Skip this combination.
                logger.warning(
                    f"Internal memory state {cell_type.internal_state} is not observable from any output "
                    f"when {other_pin_values}. Skipping input combination.")
                continue

            # Take just one of the observer output pins.
            data_out_pin = observer_outputs[0]
            logger.debug(f"Output pin: {data_out_pin}")

            # == Start characterization ==

            # Convert from sympy.Symbol to string.
            data_in_pin = str(data_in_pin)
            data_out_pin = str(data_out_pin)

            def find_min_clock_pulse_width(clock_pulse_polarity: bool, rising_data_edge: bool):
                min_clock_pulse_width, delay = find_minimum_clock_pulse_width(
                    cell_config=cell_conf,
                    ff_clock_edge_polarity=clock_edge_polarity,
                    clock_input=clock_pin,
                    data_in=data_in_pin,
                    data_out=data_out_pin,
                    setup_time=2e-9,  # Choose a reasonably large setup time.
                    clock_pulse_polarity=clock_pulse_polarity,
                    rising_data_edge=rising_data_edge,
                    clock_rise_time=10e-12,  # TODO: Something fails when 0.
                    clock_fall_time=10e-12,  # TODO: How to choose this?
                    output_load_capacitances={data_out_pin: 0},
                    clock_pulse_width_guess=100e-12,
                    max_delay_estimation=1e-7,
                    static_input_voltages=static_input_voltages,
                )
                logger.debug(f'min_clock_pulse_width = {min_clock_pulse_width}, delay = {delay}')
                return min_clock_pulse_width, delay

            # Find the minimal clock pulse for negative and positive pulses.
            # For each pulse type inspect rising and falling data edges.
            logger.info(f"Find minimal clock pulse width ({clock_pin}).")
            min_pulse_width_low, _delay = max(find_min_clock_pulse_width(False, False),
                                              find_min_clock_pulse_width(False, True))
            logger.info(f"min_pulse_width_low = {min_pulse_width_low} s")
            min_pulse_width_high, _delay = max(find_min_clock_pulse_width(True, False),
                                               find_min_clock_pulse_width(True, True))
            logger.info(f"min_pulse_width_high = {min_pulse_width_high} s")
            
            # Find the shortest clock pulse width which samles a stable data signal for all cases.
            clock_pulse_width_guess = max(clock_pulse_width_guess, min_pulse_width_low, min_pulse_width_high)

            # Write information on clock pin to liberty.
            # TODO: minimum pulse width is potentially computed for many different input combinations. Take min/max of them! (Now just the last one will be stored)
            clock_pin_group = new_cell_group.get_group('pin', clock_pin)
            clock_pin_group['clock'] = 'true'
            clock_pin_group['min_pulse_width_high'] = min_pulse_width_high / conf.time_unit
            clock_pin_group['min_pulse_width_low'] = min_pulse_width_low / conf.time_unit

            # Find setup and hold times.
            result = characterize_flip_flop_setup_hold(
                cell_conf=cell_conf,
                data_in_pin=data_in_pin,
                data_out_pin=data_out_pin,
                clock_pin=clock_pin,
                clock_edge_polarity=clock_edge_polarity,

                constrained_pin_transition=input_transition_times,
                related_pin_transition=related_pin_transition,

                output_load_capacitance=0,  # TODO: Is it accurate to assume zero output load?

                static_input_voltages=static_input_voltages
            )

            # Get the table indices.
            # TODO: get correct index/variable mapping from liberty file.
            index_1 = result['related_pin_transition'] / conf.time_unit
            index_2 = result['constrained_pin_transition'] / conf.time_unit
            # TODO: remember all necessary templates and create template tables.

            input_pin_group = new_cell_group.get_group('pin', data_in_pin)

            clock_edge = 'rising' if clock_edge_polarity else 'falling'

            # Add setup/hold information to the liberty pin group.
            for constraint_type in ['hold', 'setup']:
                template_table = liberty_util.create_constraint_template_table(
                    new_library, constraint_type, len(index_1), len(index_2)
                )
                table_template_name = template_table.args[0]

                rise_constraint = Group('rise_constraint', args=[table_template_name])
                rise_constraint.set_array('index_1', index_1)
                rise_constraint.set_array('index_2', index_2)
                rise_constraint.set_array(
                    'values',
                    result[f'{constraint_type}_rise_constraint'] / conf.time_unit
                )

                fall_constraint = Group('fall_constraint', args=[table_template_name])
                fall_constraint.set_array('index_1', index_1)
                fall_constraint.set_array('index_2', index_2)
                fall_constraint.set_array(
                    'values',
                    result[f'{constraint_type}_fall_constraint'] / conf.time_unit
                )

                timing_group = Group(
                    'timing',
                    attributes=[
                        Attribute('timing_type', f'{constraint_type}_{clock_edge}'),
                        Attribute('related_pin', EscapedString(clock_pin))
                    ],
                    groups=[rise_constraint, fall_constraint]
                )

                if len(when_other_inputs.atoms(sympy.Symbol)) > 0:
                    timing_group.set_boolean_function('when', when_other_inputs)

                input_pin_group.groups.append(timing_group)

    
    # Measure clock-to-output delays.
    for data_out_pin in data_out_pins:
        
        logger.info(f"Measure clock-to-output delay for output '{data_out_pin}'")
        
        output_function = cell_type.outputs[data_out_pin].function
        logger.info(f"Output function: {output_function}")
        
        # Convert output pin to string
        data_out_pin_name = str(data_out_pin)
    
        output_pin_group = new_cell_group.get_group('pin', data_out_pin_name)
        output_pin_group.set_boolean_function('function', output_function)
        related_pin = clock_pin
        
        # Find data input pins which can control the output pin.
        possible_input_pins = find_flipflop_state_control_and_observe_pins(cell_type, preset_clear_input)

        assert len(possible_input_pins) > 0, "No input pin found which can control the flip-flop state."
            
        # Select an input pin and a combination of other input values such that the input pin controls output of the flip-flop.
        data_in_pin, static_input_values, observer_outputs = possible_input_pins[0]

        assert data_out_pin in observer_outputs, f"Flip-flop state is not observable from {data_out_pin} when {static_input_values}."

        # Translate logic values to voltages.
        static_input_voltages = {
            pin: value * cell_conf.global_conf.supply_voltage 
            for pin, value in static_input_values.items()
        }

        logger.info(f"Use input pin {data_in_pin} with other pin values {static_input_voltages}")
     
        for clock_transition_time in related_pin_transition:
            for output_capacitance in output_capacitances:
        
                output_load_capacitances = {data_out_pin_name: output_capacitance} 
                
                for data_edge_polarity in [False, True]:
                    
                    # Choose setup and hold time to be used for the delay measurement.
                    settings = {
                        CalcMode.BEST: (1e-9, 1e-9), # Use large setup and hold times. This leads to shorter propagation time.
                        CalcMode.WORST: (1e-9, 1e-9), # Use short setup and hold times. This leads to longer propagation times.
                        CalcMode.TYPICAL: (1e-9, 1e-9),
                    }
                    
                    (setup_time, hold_time) = settings[conf.timing_corner]
                    
                    clock_cycle_hint = clock_pulse_width_guess * 2 # Choose a big-enough clock pulse such that data is sampled for sure.
        
                    data_transition_time = 1e-13 # TODO
                    
                    delay, slew_time = measure_clock_to_output_delay(
                        cell_conf,
                        data_in_pin,
                        data_out_pin_name,
                        clock_pin,
                        clock_edge_polarity,
                        data_edge_polarity,
                        output_load_capacitances,
                        clock_transition_time,
                        data_transition_time,
                        setup_time,
                        hold_time,
                        clock_cycle_hint,
                        static_input_voltages
                    )
                    print(f"delay = {delay}")
        
        #timing_group = Group(
        #    'timing',
        #    attributes={
        #        'timing_type': [f'rising_edge'],
        #        'related_pin': [EscapedString(related_pin)]
        #    },
        #    groups=[cell_rise, cell_fall]
        #)

def find_flipflop_state_control_and_observe_pins(
    cell_type: SingleEdgeDFF,
    other_inputs: Dict[sympy.Symbol, sympy.Symbol]
) -> List[Tuple[str, Dict[str, bool]]]:
    """
    Find data input pins which control the state of the flip-flop.


    Returns a list of pins which can control the flip-flop state together with 
    1) an assignment of other input values (clear, preset, ...) such that the input pin becomes controlling
    2) a list of output pins which allow to observe the internal state with the given other input values
    """

    assert isinstance(cell_type, SingleEdgeDFF), "cell_type must be a SingleEdgeDFF"
    
    # Find all data pins that are relevant for the internal state of the flip-flop.
    data_in_pins = sorted(cell_type.next_state.atoms(sympy.Symbol))
    logger.debug(f"Input pins relevant for internal state: {data_in_pins}")

    # Find data input pins which can control the output pin.
    controlling_input_pins = []
    for data_in_pin in data_in_pins:
    
        # Find all assignments of the other data pins such that the data pin controls
        # the internal state.
        # Find values of the other pins such that:
        #  next_state(data_in_pin=0, other_pins) != next_state(data_in_pin=1, other_pins)
        next_state_0 = cell_type.next_state.subs({data_in_pin: False})
        next_state_1 = cell_type.next_state.subs({data_in_pin: True})
        models = list(satisfiable(next_state_0 ^ next_state_1, all_models=True))

        for other_pin_values in models:
            # Express the assignment of the other pins as a boolean formula.
            # This will also be used as a 'when' statement in the liberty file.
            when_other_inputs = sympy.And(*(pin if value else ~pin for pin, value in other_pin_values.items()))
            logger.info(f"Measure clock-to-output delay with data input {data_in_pin} when {when_other_inputs}.")

            # Set static voltages of other input pins.
            other_pin_values.update(other_inputs)
            
            logger.debug(f"Pin {data_in_pin} controls the flip-flop state when: {other_pin_values}")
            

            static_input_voltages = dict()
            for pin, value in other_pin_values.items():
                if not isinstance(pin, sympy.Symbol):
                    continue
                value = value == True
                pin = str(pin)
                logger.debug(f"{pin} = {value} V")
                static_input_voltages[pin] = value
                
            # Find an output pin such that the internal state is observable.
            observer_outputs = []  # Output pins that can observe the internal memory state.
            for output_pin, output in cell_type.outputs.items():
                function = output.function
                # Substitute with constant input pins.
                function = function.subs(other_pin_values)

                # Compute the output for all values of the internal state and make sure it is different.
                function0 = function.subs({cell_type.internal_state: False})
                function1 = function.subs({cell_type.internal_state: True})
                is_observable = not satisfiable(~(function0 ^ function1))  # Test if function0 != function1

                logger.debug(f"Internal state {cell_type.internal_state} observable from output {output_pin} "
                             f"when {other_pin_values}: {is_observable}")

                if is_observable:
                    observer_outputs.append(output_pin)

            logger.debug(f"Internal state is observable from: {observer_outputs}")

            # Remember the pin which controls the internal state and the the required
            # voltages at the other pins.
            controlling_input_pins.append((str(data_in_pin), static_input_voltages, observer_outputs))

    return controlling_input_pins





# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from PySpice.Probe.Plot import plot
from PySpice.Probe.WaveForm import TransientAnalysis
from PySpice.Spice.Netlist import Circuit
from PySpice.Spice.NgSpice.Shared import NgSpiceShared
from PySpice.Spice.Parser import SpiceParser
from PySpice.Unit import *
from PySpice.Unit.SiUnits import Farad, Second, Volt
from typing import Any, Dict, Union, Optional

from .piece_wise_linear import PieceWiseLinear

logger = logging.getLogger(__name__)


def simulate_circuit(circuit: Circuit,
                     input_voltages: Dict[Any, Union[float, PieceWiseLinear]],
                     step_time: Second,
                     end_time: Second,
                     temperature: float = 25,
                     initial_voltages: Optional[Dict[str, Volt]] = None,
                     simulator='ngspice-shared'):
    """
    Simulate a circuit with given input voltages.
    :param circuit:
    :param input_voltages: Dict[node name, input voltage]
        Input voltages can either be a `float` for constant values or a `.util.PieceWiseLinearWave`
        changing wave forms.
    :param step_time: Time step of simulation.
    :param end_time: End time of simulation.
    :param temperature: Simulation temperature.
    :param initial_voltages: Dict[node name, voltage].
        Initial voltages for nodes. Default = None.
    :param simulator: Choose the simulator.
        One of {'ngspice-shared', 'ngspice-subprocess', 'xyce-serial', 'xyce-parallel'}
    :return: PySpice Analysis object.
    """
    circuit = circuit.clone()

    logger.info("Instantiate ngspice simulator.")
    # ngspice_shared = NgSpiceShared.new_instance(send_data=False)
    # simulator = circuit.simulator(temperature=temperature,
    #                               nominal_temperature=temperature,
    #                               simulator='ngspice-shared',
    #                               ngspice_shared=ngspice_shared
    #                               )

    simulator = circuit.simulator(temperature=temperature,
                                  nominal_temperature=temperature,
                                  simulator=simulator
                                  )

    # Create input drivers.
    # This will be either constant voltages for static input signals or piece-wise linear sources for dynamic inputs.
    logger.debug("Create input drivers.")
    for name, voltage in input_voltages.items():
        if isinstance(voltage, PieceWiseLinear):
            piece_wise_linear_voltage_source(circuit, 'in_{}'.format(name), name, circuit.gnd,
                                             voltage
                                             )
            # simulator.initial_condition(**{name: voltage(0) @ u_V})
        else:
            # simulator.initial_condition(**{name: voltage @ u_V})
            circuit.V('in_{}'.format(name), name, circuit.gnd, voltage @ u_V)

    if initial_voltages is not None:
        simulator.initial_condition(**initial_voltages)

    # Run transient analysis.
    # Set use_initial_condition (uic) to False to enable DC bias computation. See ngspice manual 15.2.2 2)
    logger.info("Run transient analysis. step_time = {}, end_time = {}".format(step_time, end_time))

    analysis = simulator.transient(step_time=step_time,
                                   end_time=end_time,
                                   use_initial_condition=False
                                   )
    logger.info("Transient analysis done.")
    return analysis


def piece_wise_linear_voltage_source(circuit: Circuit, name: str, plus, minus, wave: PieceWiseLinear,
                                     repeat=None, time_delay=None):
    """ Create a piece wise linear voltage source.
    This is a helper function needed because PWL sources are not properly handled by PySpice at time of this writing.

    :param circuit: The netlist to add the source.
    :param name: Name of the voltage source.
    :param plus: Positive terminal.
    :param minus: Negative terminal.
    :param times: Time axis.
    :param voltages: Voltage axis.
    :param repeat: Number of times the sequence should be repeated. Default = 1. 0 means endless repetition.
    :param time_delay: Delay the signal by this amount of time. Default = 0.
    :return: Return a handle to the generated voltages source.
    """
    pwl_string = ' '.join((
        '%es %eV' % (float(time), float(voltage))
        for time, voltage in zip(wave.x, wave.y)
    ))

    pwl_args = []
    if repeat is not None:
        pwl_args.append('r={}'.format(repeat))
    if time_delay is not None:
        pwl_args.append('td={}'.format(float(time_delay)))

    return circuit.V(name, plus, minus,
                     'PWL({}) DC=0'.format(' '.join([pwl_string] + pwl_args))
                     )



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

name = "lctime"

import sys
import os

def trace_calls(frame, event, arg):
    if event != 'call':
        return
    co = frame.f_code
    func_name = co.co_name
    if func_name == 'write':
        # Ignore write() calls from print statements
        return
    func_line_no = frame.f_lineno
    func_filename = co.co_filename
    caller = frame.f_back
    if caller is None:
        print('Call to %s on line %s of %s from unknown' % \
          (func_name, func_line_no, func_filename))
    else:
        caller_line_no = caller.f_lineno
        caller_filename = caller.f_code.co_filename
        print('Call to %s on line %s of %s from line %s of %s' % \
          (func_name, func_line_no, func_filename,
          caller_line_no, caller_filename))
    return

if 'DEBUG_LCTIME' in os.environ:
    sys.settrace(trace_calls)



# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from pysmt.shortcuts import *

# Define 4-valued logic type.
#TL = Type('L')
#TH = Type('H')
#TX = Type('X')
#TZ = Type('Z')

FVLType = Type('FVL', arity=2)
# FVL = Type('FVL', arity=0)
# FVL = INT

env = get_env()
FVL = env.type_manager.get_type_instance(FVLType, BOOL, BOOL)
#L = env.type_manager.get_type_instance(FVL, TL)
#H = env.type_manager.get_type_instance(FVL, TH)
#X = env.type_manager.get_type_instance(FVL, TX)
#Z = env.type_manager.get_type_instance(FVL, TZ)

# Low
L = Symbol('L', FVL)
# High
H = Symbol('H', FVL)
# Unknown
X = Symbol('X', FVL)
# High-impedance
Z = Symbol('Z', FVL)


# Helper shortcut
Eq = EqualsOrIff

# Define boolean functions.
f_not = Symbol('f_not', FunctionType(FVL, [FVL]))

definitions = [
        # L,H,X,Z are all different.
        Not(Eq(L, H)),
        Not(Eq(L, X)),
        Not(Eq(L, Z)),
        Not(Eq(H, X)),
        Not(Eq(H, Z)),
        Not(Eq(X, Z)),

        # Define L,H,X,Z
        # Eq(L, Int(0)),
        # Eq(H, Int(1)),
        # Eq(X, Int(2)),
        # Eq(Z, Int(3)),

        # Define 'not' function.
        Eq(f_not(L), H),
        Eq(f_not(H), L),
        Eq(f_not(X), X),
        Eq(f_not(Z), X),
        ]

solver = Solver()
for d in definitions:
    solver.add_assertion(d)

print(FVL.as_smtlib())

x = Symbol('x', FVL)
solver.add_assertion(Eq(x, f_not(x)))
solver.add_assertion(Not(Eq(x, L)))
solver.add_assertion(Not(Eq(x, H)))
solver.add_assertion(Not(Eq(x, X)))
solver.add_assertion(Not(Eq(x, Z)))
# solver.add_assertion(Eq(x, f_not(f_not(H))))
# solver.add_assertion(Not(Eq(x, X)))
# solver.add_assertion(Not(Eq(x, Z)))
sat = solver.check_sat()

print('sat =', sat)
if sat:
        model = solver.get_model()
        # print(model)
        # print(model[x])



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from PySpice.Spice.Netlist import Circuit
from PySpice.Spice.NgSpice.Shared import NgSpiceShared
from PySpice.Spice.Parser import SpiceParser
from PySpice.Unit import *
from PySpice.Unit.SiUnits import Farad, Second

import PySpice.Logging.Logging as Logging
import numpy as np

from .ngspice_simulation import piece_wise_linear_voltage_source
from .piece_wise_linear import PieceWiseLinear

import matplotlib.pyplot as plt


# def test_simulate_circuit():
#     time_step = 1 @ u_ms
#     end_time = 10 @ u_s
#
#     circuit = Circuit(title='test1')
#
#     # circuit.V('VDD', 'vdd', circuit.gnd, 10 @ u_V)
#     circuit.R('1', 'input', 'a', 1 @ u_kOhm)
#     circuit.C('1', circuit.gnd, 'a', 1000 @ u_uF)
#
#     ngspice_shared = NgSpiceShared.new_instance(send_data=False)
#
#     piece_wise_linear_voltage_source(circuit, 'src', 'input', circuit.gnd,
#                                      PieceWiseLinear([0, 1, 2, 10], [0, 1, 1, 1]))
#
#     # simulator = circuit.simulator(temperature=25,
#     #                               nominal_temperature=25,
#     #                               simulator='ngspice-shared',
#     #                               ngspice_shared=ngspice_shared)
#
#     ngspice_shared.stop('v(a) > 0.5')
#     # ngspice_shared.destroy()
#     # print(str(simulator))
#     # ngspice_shared.load_circuit(str(simulator))
#     # ngspice_shared.run()
#     # analysis = ngspice_shared.plot(simulator, 'asdf').to_analysis()
#
#     simulator = circuit.simulator(temperature=25,
#                                   nominal_temperature=25,
#                                   simulator='ngspice-shared',
#                                   ngspice_shared=ngspice_shared)
#     analysis = simulator.transient(step_time=time_step, end_time=end_time)
#
#     print(np.array(analysis['a']))
#
#     # plt.plot(analysis['a'])
#     # plt.plot(analysis['input'])
#     # plt.show()


# def test_simulate_circuit_lowlevel_control():
#     """
#     Use `NgSpiceShared` to get finer grained control over the simulation.
#     Especially needed for breakpoints.
#     """
#     circuit = Circuit(title='test2')
#
#     circuit.R('1', 'input', 'a', 1 @ u_kOhm)
#     circuit.C('1', circuit.gnd, 'a', 1000 @ u_uF)
#
#     # piece_wise_linear_voltage_source(circuit, 'src', 'input', circuit.gnd,
#     #                                  PieceWiseLinear([0, 1, 2, 10], [0, 1, 1, 0]))
#     circuit.V('src', 'input', circuit.gnd, 0)
#
#     ngs = NgSpiceShared.new_instance(send_data=False)
#
#     ngs.destroy()
#     ngs.remove_circuit()
#
#     netlist = "{}\n.end".format(str(circuit))
#     print(netlist)
#     ngs.load_circuit(netlist)
#     # Change capacitance.
#     # ngs.alter_device('C1', C="1000uF")
#     # Change signal of piece wise linar voltage source..
#     wave = PieceWiseLinear([0, 1, 10], [0, 1, 1])
#     pwl_string = ' '.join((
#         '%es %eV' % (float(time), float(voltage))
#         for time, voltage in zip(wave.x, wave.y)
#     ))
#     ngs.alter_device('Vsrc', PWL="({})".format(pwl_string))
#
#     # Set temperature.
#     ngs.option(temp=25)
#     ngs.option(nomtemp=25)
#     # Set breakpoint.
#     ngs.stop('v(a) > 0.9')
#     # Run simulation.
#     ngs.exec_command("tran 100ms 60s")
#     ngs.run(background=False)
#
#     # Retreive signals.
#     plot_name = ngs.last_plot
#     print(plot_name)
#     if plot_name == 'const':
#         raise Exception("Simulation failed.")
#     plot = ngs.plot(None, plot_name)
#     analysis = plot.to_analysis()
#
#     print(np.array(analysis['a']))
#     plt.plot(analysis.time, analysis['input'], 'x-')
#     plt.plot(analysis.time, np.array(analysis['a']), 'x-')
#     plt.show()


def test3():

    netlist = r""".title Input capacitance measurement for pin "A"
.include /home/kramert/unison/local/blackbird/git/librecell-examples/librecell_invx1_example/gpdk45nm.m
.include /home/kramert/unison/local/blackbird/git/librecell-examples/librecell_invx1_example/INVX1.pex.netlist
Xcircuit_unter_test A GND VDD Y INVX1
Vpower_vdd VDD GND 1.1V
Isrc_A GND A 10000nA

.end

"""
    ngs = NgSpiceShared.new_instance(send_data=False)

    ngs.destroy()
    ngs.remove_circuit()

    print(netlist)
    ngs.load_circuit(netlist)

    # Run simulation.
    ngs.exec_command("tran 1ns 100ns")
    ngs.run(background=False)

    # Retreive signals.
    plot_name = ngs.last_plot
    print(plot_name)
    if plot_name == 'const':
        raise Exception("Simulation failed.")
    plot = ngs.plot(None, plot_name)
    analysis = plot.to_analysis()

    plt.plot(analysis.time, analysis['y'], 'x-')
    plt.plot(analysis.time, np.array(analysis['a']), 'x-')
    plt.show()


# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Tool for merging liberty libraries.
"""

from liberty.parser import parse_liberty
from liberty.types import Group
from typing import List
import argparse
import logging
from itertools import chain
from copy import copy
import os

logger = logging.getLogger(__name__)


def add_named_group(base_library: Group, replacement: Group, replace_if_already_exists: bool = False) -> None:
    """
    Add or replace a named group in the base library. The name of the group is `group.args[0]`.
    If the group does not yet exist it will be appended to the base library. If it exists, then
    depending on `replace_if_already_exists` the existing group will be replaced with the `replacement`.
    :param base_library: Group container to be updated.
    :param replacement:
    :param replace_if_already_exists:
    :return: None
    """
    cell_name = replacement.args[0]

    for i in range(len(base_library.groups)):
        g = base_library.groups[i]
        if g.group_name == replacement.group_name and g.args[0] == cell_name:
            if replace_if_already_exists:
                logger.info("Replace group: {}({})".format(replacement.group_name, replacement.args[0]))
                base_library.groups[i] = replacement
            return

    # If cell has not been replaced append the new cell.
    logger.info("Add group: {}({})".format(replacement.group_name, replacement.args[0]))
    base_library.groups.append(replacement)


def add_named_groups(base_library: Group, replacements: List[Group], replace_if_already_exists: bool = False) -> None:
    for repl in replacements:
        add_named_group(base_library, repl, replace_if_already_exists)


def main():
    """
    Command-line for merging liberty libraries.
    :return:
    """

    logger = logging.getLogger(__name__)

    parser = argparse.ArgumentParser(description='Merge liberty libraries. The content of the base library will'
                                                 'be replaced by the content in the update libraries.')

    parser.add_argument('-b', '--base', required=True, metavar='LIBERTY_BASE', type=str, help='Base liberty file.')

    parser.add_argument('-o', '--output', required=True, metavar='LIBERTY_OUT', type=str, help='Output liberty file.')

    parser.add_argument('-u', '--update', required=False, action='append', nargs='+', metavar='LIBERTY', type=str,
                        help='Liberty files with updates.')

    parser.add_argument('--debug', action='store_true', help='Enable debug mode.')
    parser.add_argument('--force', action='store_true', help='Allow overwriting of output file.')

    merge_modes = {'replace_cells'}

    parser.add_argument('--merge', default='replace_cells',
                        required=False, choices=merge_modes,
                        metavar='MERGE_MODE', type=str, help='Choose how to merge.')

    # Parse arguments
    args = parser.parse_args()

    DEBUG = args.debug
    log_level = logging.DEBUG if DEBUG else logging.INFO

    logging.basicConfig(format='%(levelname)8s: %(message)s', level=log_level)

    # Check if output would overwrite something.
    if not args.force and os.path.exists(args.output):
        logger.error("Output file exists. Use --force to overwrite it.")
        exit(1)

    # Read base liberty.
    base_lib_file = args.base
    with open(base_lib_file) as f:
        logger.info("Reading base liberty: {}".format(base_lib_file))
        data = f.read()
    base_library = parse_liberty(data)

    # Read updates
    update_libraries = []
    for lib_file in chain(*args.update):
        with open(lib_file) as f:
            logger.info("Reading liberty: {}".format(lib_file))
            data = f.read()
        lib = parse_liberty(data)
        update_libraries.append(lib)

    if args.merge == 'replace_cells':

        new_library = copy(base_library)

        # Add updates to new library.
        for lib in update_libraries:
            # Add cell groups
            add_named_groups(new_library, lib.get_groups('cell'), replace_if_already_exists=True)

            # Add non-existent lookup table groups
            add_named_groups(new_library, lib.get_groups('lu_table_template'), replace_if_already_exists=False)
            add_named_groups(new_library, lib.get_groups('power_lut_template'), replace_if_already_exists=False)

        num_cells_old = len(base_library.get_groups('cell'))
        num_cells_new = len(new_library.get_groups('cell'))

        logger.info("Number of cells in base: {}, number of cells in output: {}".format(num_cells_old, num_cells_new))

        with open(args.output, 'w') as f:
            logger.info('Write liberty: {}'.format(args.output))
            f.write(str(new_library))



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Helpers related to licencing.
"""


def original_source_repository() -> str:
    return "https://codeberg.org/librecell/lctime"


def licence_notice_string_single_line() -> str:
    return "This program is licenced under the AGPL-3.0-or-later licence. Source code:  {}".format(
        original_source_repository())



# Copyright (c) 2021 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from .functional_abstraction import *

import itertools
import networkx as nx
from typing import Any, Dict, List, Iterable, Tuple, Set, Optional, Union
from enum import Enum
import collections
import sympy
from sympy.logic import satisfiable, simplify_logic as sympy_simplify_logic
from sympy.logic import boolalg

import logging

from ..cell_types import Latch, SingleEdgeDFF

# logging.basicConfig(level=logging.INFO, stream=sys.stdout)
logger = logging.getLogger(__name__)

"""
Recognize sequential cells based on the output of the `functional_abstraction.analyze_circuit_graph()` function.

For each class of cells (latch, single-edge triggered flip-flop, ...) a Extractor class should be created.
The extractor class tries to recognize the cell from an abstract formal description.
For recognizing an unknown cell, all extractor classes are tried until one finds a match.
"""


# def find_clear_and_preset_signals(f: boolalg.Boolean) -> Dict[boolalg.BooleanAtom, Tuple[bool, bool]]:
#     """
#     Find the variables in a boolean formula that can either force the formula to True or False.
#     :param f:
#     :return: Dict[Variable symbol, (is preset, is active high)]
#     """
#     results = dict()
#     atoms = f.atoms(sympy.Symbol)
#     for a in atoms:
#         for v in [False, True]:
#             f.subs({a: v})
#             if f == sympy.true and f != sympy.false:
#                 results[a] = (True, v)
#             elif f == sympy.false and f != sympy.true:
#                 results[a] = (False, v)
#     return results
#
# def test_find_clear_and_preset_signals():
#     a = sympy.Symbol('a')
#     clear = sympy.Symbol('clear')
#     preset = sympy.Symbol('preset')

def find_boolean_isomorphism(a: boolalg.Boolean, b: boolalg.Boolean) -> Optional[
    Dict[boolalg.BooleanAtom, boolalg.BooleanAtom]]:
    """
    Find a one-to-one mapping of the variables in `a` to the variables in `b` such that the both formulas are
    equivalent. Return `None` if there is no such mapping.
    The mapping is not necessarily unique.
    :param a:
    :param b:
    :return:
    """

    a_vars = list(a.atoms(sympy.Symbol))
    b_vars = list(b.atoms(sympy.Symbol))

    if len(a_vars) != len(b_vars):
        return None

    # Do a brute-force search.
    for b_vars_perm in itertools.permutations(b_vars):
        substitution = {old: new for old, new in zip(a_vars, b_vars_perm)}
        a_perm = a.subs(substitution)
        # Check for structural equality first, then for mathematical equality.
        if a_perm == b or bool_equals(a_perm, b):
            return substitution

    return None


def test_find_boolean_isomorphism():
    assert find_boolean_isomorphism(sympy.true, sympy.true) is not None
    assert find_boolean_isomorphism(sympy.true, sympy.false) is None

    a, b, c, x, y, z = sympy.symbols('a b c x y z')
    f = (a & b) | c
    g = (x & y) | z
    mapping = find_boolean_isomorphism(f, g)
    assert mapping == {a: x, b: y, c: z}

    # MUX
    f = (a & c) | (b & ~c)
    g = ~((~x & z) | (~y & ~z))
    mapping = find_boolean_isomorphism(f, g)
    print(mapping)
    assert mapping == {a: x, b: y, c: z}


class LatchExtractor:
    def __init__(self):
        pass

    def extract(self, c: AbstractCircuit) -> Optional[Latch]:
        """
        Try to recognize a latch based on the abstract circuit representation.
        :param c:
        :return:
        """
        logger.debug("Try to extract a latch.")
        if len(c.latches) != 1:
            logger.debug(f"Not a latch. Wrong number of latches. Need 1, found {len(c.latches)}.")
            return None

        output_nets = list(c.output_pins)
        if len(output_nets) != 1:
            logger.debug(f"Expect 1 output net, found {len(output_nets)}.")
            return None

        # Trace back the output towards the inputs.
        output_net = output_nets[0]
        output = c.outputs[output_net]

        # Check that the output is not tri-state.
        if output.is_tristate():
            logger.debug("Can't recognize DFF with tri-state output.")
            return None

        # Trace back the output towards the inputs.
        latch_path = []
        current_nodes = set(output.function.atoms(sympy.Symbol))
        while current_nodes:
            node = current_nodes.pop()
            if node in c.outputs:
                next = c.outputs[node]
                current_nodes.update(next.function.atoms(sympy.Symbol))
            elif node in c.latches:
                latch = c.latches[node]
                assert isinstance(latch, Memory)

                latch_path.append(latch)
                current_nodes = set(latch.data.atoms(sympy.Symbol))
            else:
                # Cannot further resolve the node, must be an input.
                logger.debug(f"Input node: {node}")

        if len(latch_path) != 1:
            logger.debug(f"No latch found in the fan-in tree of the outputs {output.function.atoms()}.")
            return None

        latch = latch_path[0]

        # Find preset / clear conditions.
        enable_signals = sorted(latch.write_condition.atoms(sympy.Symbol), key=lambda s: s.name)
        logger.debug(f"Potential clock/set/preset signals {enable_signals}")

        # Find preset condition.
        preset_condition = simplify_logic(latch.data & latch.write_condition)
        logger.info(f"Preset condition: {preset_condition}")

        # Find clear condition.
        clear_condition = simplify_logic(~latch.data & latch.write_condition)
        logger.info(f"Clear condition: {clear_condition}")

        # Find single signals that satisfy the preset.
        preset_signals = [s for s in enable_signals if bool_equals(preset_condition.subs({s: True}), sympy.true)]
        preset_signals_inv = [s for s in enable_signals if bool_equals(preset_condition.subs({s: False}), sympy.true)]
        logger.info(f"Preset signals (active high): {preset_signals}")
        logger.info(f"Preset signals (active low): {preset_signals_inv}")

        # Find single signals that satisfy the clear condition.
        clear_signals = [s for s in enable_signals if bool_equals(clear_condition.subs({s: True}), sympy.true)]
        clear_signals_inv = [s for s in enable_signals if bool_equals(clear_condition.subs({s: False}), sympy.true)]
        logger.info(f"Clear signals (active high): {clear_signals}")
        logger.info(f"Clear signals (active low): {clear_signals_inv}")

        result = Latch()
        result.outputs = {output_net: output}
        result.enable = latch.write_condition
        result.data_in = latch.data
        result.clear = clear_condition
        result.preset = preset_condition

        return result


class DFFExtractor:
    def __init__(self):
        pass

    def extract(self, c: AbstractCircuit) -> Optional[SingleEdgeDFF]:
        """
        Try to recognize a single-edge triggered D-flip-flop based on the abstract circuit representation.
        :param c:
        :return:
        """
        logger.debug("Try to extract a D-flip-flop.")
        if len(c.latches) != 2:
            logger.debug(f"Not a flip-flop. Wrong number of latches. Need 2, found {len(c.latches)}.")
            return None

        output_nets = list(c.output_pins)
        if len(output_nets) not in [1, 2]:
            logger.debug(f"Expect 1 or 2 output nets, found {len(output_nets)}.")
            return None

        # Trace back the output towards the inputs.
        outputs = [c.outputs[n] for n in output_nets]

        # Check that the output is not tri-state.
        for output in outputs:
            if output.is_tristate():
                logger.warning(f"'{output}' is a tri-state output.")
                logger.warning("Can't recognize DFF with tri-state output.")
                return None

        # Trace back the output towards the inputs.
        latch_path = []
        current_nodes = set()
        for output in outputs:
            current_nodes.update(output.function.atoms(sympy.Symbol))
        while current_nodes:
            node = current_nodes.pop()
            if node in c.outputs:
                next = c.outputs[node]
                current_nodes.update(next.function.atoms(sympy.Symbol))
            elif node in c.latches:
                latch = c.latches[node]
                assert isinstance(latch, Memory)

                latch_path.append((latch, node))
                current_nodes = set(latch.data.atoms(sympy.Symbol))
            else:
                # Cannot further resolve the node, must be an input.
                logger.debug(f"Input node: {node}")

        # Find trigger condition of the flip-flop.
        latch1, latch1_output_node = latch_path[1]
        latch2, latch2_output_node = latch_path[0]

        # Find condition such that the FF is in normal operation mode.
        # I.e. no set or reset is active.
        ff_normal_condition = simplify_logic(latch1.write_condition ^ latch2.write_condition)
        logger.debug(f"FF normal-mode condition: {ff_normal_condition}")

        # Find clock inputs of the latches.
        latch1_write_normal = simplify_with_assumption(assumption=ff_normal_condition, formula=latch1.write_condition)
        latch2_write_normal = simplify_with_assumption(assumption=ff_normal_condition, formula=latch2.write_condition)

        logger.debug(f"Latch clocks: {latch1_write_normal}, {latch2_write_normal}")

        clocked_on = simplify_logic(~latch1_write_normal & latch2_write_normal)
        logger.debug(f"clocked_on: {clocked_on}")

        clock_signals = set(latch1_write_normal.atoms(sympy.Symbol)) | set(latch2_write_normal.atoms(sympy.Symbol))
        if len(clock_signals) != 1:
            logger.warning(f"Clock must be a single signal. Found: {clock_signals}")
            return None

        clock_signal = clock_signals.pop()
        logger.info(f"Clock signal: {clock_signal}")

        assert not satisfiable(boolalg.Equivalent(latch1_write_normal, latch2_write_normal)), \
            "Clock polarities of latches must be complementary."

        # Find polarity of the active clock-edge.
        active_edge_polarity = latch2_write_normal.subs({clock_signal: True})

        # Sanity check: The both latches must be transparent for the opposite phases of the clock.
        assert active_edge_polarity == latch1_write_normal.subs({clock_signal: False}), \
            "Latches must be transparent in opposite phases of the clock."

        logger.info(f"Active edge polarity: {'rising' if active_edge_polarity else 'falling'}")

        # Assemble D-flip-flop description object.
        dff = SingleEdgeDFF()
        dff.clocked_on = clocked_on
        # dff.clock_edge_polarity = active_edge_polarity

        # == Detect asynchronous set/reset signals ==
        potential_set_reset_signals = list((set(latch1.write_condition.atoms(sympy.Symbol))
                                            | set(latch2.write_condition.atoms(sympy.Symbol))) - {clock_signal}
                                           )
        logger.debug(
            f"Potential asynchronous set/reset signals: {sorted(potential_set_reset_signals, key=lambda n: n.name)}")

        if potential_set_reset_signals:
            # More than 2 asynchronous set/reset signals are not supported.
            if len(potential_set_reset_signals) > 2:
                logger.error(f"Cannot recognize flip-flops with more than 2 asynchronous set/reset signals "
                             f"(found {potential_set_reset_signals}).")
                return None

            # Find signal values such that the FF is in normal operation mode.
            inactive_set_reset_models = list(satisfiable(ff_normal_condition, all_models=True))
            if len(inactive_set_reset_models) != 1:
                logger.warning(f"There's not exactly one signal assignment such that the FF is in normal mode: "
                               f"{inactive_set_reset_models}")
            assert inactive_set_reset_models, f"Normal operation condition is not satisfiable: {ff_normal_condition}."

            # Determine polarity of the set/reset signals.
            sr_disabled_values = inactive_set_reset_models[0]
            logger.info(f"Set/reset disabled when: {sr_disabled_values}")
            sr_enabled_values = {k: not v for k, v in sr_disabled_values.items()}
            logger.info(f"Set/reset enabled when: {sr_enabled_values}")

            async_set_signals = []
            async_reset_signals = []

            # Find if the signals are SET or RESET signals.
            # Set just one to active and observe the flip-flop output.
            for signal in potential_set_reset_signals:
                one_active = sr_disabled_values.copy()
                signal_value = sr_enabled_values[signal]
                one_active.update({signal: signal_value})

                wc2 = latch2.write_condition.subs(one_active)
                data2 = latch2.data.subs(one_active)
                assert bool_equals(wc2, True)
                if data2:
                    logger.info(f"{signal} is an async SET/PRESET signal, active {'high' if signal_value else 'low'}.")
                    async_set_signals.append((signal, signal_value))
                else:
                    logger.info(f"{signal} is an async RESET/CLEAR signal, active {'high' if signal_value else 'low'}.")
                    async_reset_signals.append((signal, signal_value))

            # TODO: Find out what happens when all asynchronous set/reset signals are active at the same time.
            all_sr_active_wc = latch2.write_condition.subs(sr_enabled_values)
            all_sr_active_data = latch2.data.subs(sr_enabled_values)
            # TODO

            if len(async_set_signals) > 1:
                logger.error(f"Multiple async SET/PRESET signals: {async_set_signals}")
            if len(async_reset_signals) > 1:
                logger.error(f"Multiple async RESET/CLEAR signals: {async_reset_signals}")

            # Store the results in the flip-flop description object.
            if async_set_signals:
                async_preset, async_set_polarity = async_set_signals[0]
                # TODO: This can be simplified by just finding the 'preset' condition.
                if async_set_polarity:
                    dff.async_preset = async_preset
                else:
                    dff.async_preset = ~async_preset
            if async_reset_signals:
                async_clear, async_reset_polarity = async_reset_signals[0]
                if async_reset_polarity:
                    dff.async_clear = async_clear
                else:
                    dff.async_clear = ~async_clear

        # Eliminate Set/Reset.
        latch1_data_normal = simplify_with_assumption(ff_normal_condition, latch1.data)
        latch2_data_normal = simplify_with_assumption(ff_normal_condition, latch2.data)
        # Eliminate clock.
        latch1_data_normal = latch1_data_normal.subs({clock_signal: not active_edge_polarity})
        latch2_data_normal = latch2_data_normal.subs({clock_signal: active_edge_polarity})

        ff_internal_state_name = sympy.Symbol('IQ')
        ff_internal_state_node = latch2_output_node

        # Resolve the data which gets written into the internal state on a clock edge.
        # Find data that gets written into the flip-flop in normal operation mode.
        ff_data_next = simplify_with_assumption(ff_normal_condition, ff_internal_state_node)

        # Resolve through second latch.
        ff_data_next = ff_data_next.subs({
            latch2_output_node: latch2_data_normal,
            clock_signal: active_edge_polarity
        })

        # Resolve through first latch.
        ff_data_next = ff_data_next.subs({
            latch1_output_node: latch1_data_normal,
            clock_signal: not active_edge_polarity
        })
        ff_data_next = simplify_logic(ff_data_next)

        dff.internal_state = ff_internal_state_name

        # Normalize by removing inversions.
        if type(ff_data_next) == boolalg.Not:
            ff_data_next = ~ff_data_next
            ff_internal_state_name = ~ff_internal_state_name

        # Store the formula of the next state.
        dff.next_state = ff_data_next

        logger.debug(f"Flip-flop output data: {dff.next_state}")

        # Find combinational output formulas. They may depend on the internal state of the flip-flop.
        ff_outputs = dict()
        for output_net in output_nets:

            # Find combinational function of the output net.
            # It should be a function of the primary inputs and of latch outputs.
            output_function = c.outputs[output_net].function
            high_impedance = c.outputs[output_net].high_impedance

            # Find data that gets written into the flip-flop in normal operation mode.
            output_formula = simplify_with_assumption(ff_normal_condition, output_function)

            # Resolve through second latch.
            output_formula = output_formula.subs({
                ff_internal_state_node: ff_internal_state_name
            })
            high_impedance = high_impedance.subs({
                ff_internal_state_node: ff_internal_state_name
            })

            logger.debug(f"Output: {output_net} = {output_formula}")
            comb = CombinationalOutput(function=output_formula, high_impedance=high_impedance)
            ff_outputs[output_net] = comb

        # Store the output functions.
        dff.outputs = ff_outputs

        # if len(ff_output_data) == 1:
        #     # FF has only one output.
        #     logger.info(f"{output_nets[0]} = {ff_output_data[0]}")
        #     out_data = ff_output_data[0]
        #     dff.data_in = out_data
        #     dff.data_out = output_nets[0]
        # elif len(ff_output_data) == 2:
        #     # Check if the outputs are inverted versions.
        #
        #     out1, out2 = output_nets
        #     out1_data, out2_data = ff_output_data
        #
        #     # Check if the outputs are inverses.
        #     if out1_data == boolalg.Not(out2_data):
        #         logger.debug("Outputs are inverses of each other.")
        #     else:
        #         logger.warning("If a flip-flop has two outputs, then need to be inverses of each other.")
        #         return None
        #
        #     # Find the inverted and non-inverted output.
        #     if type(out1_data) == boolalg.Not:
        #         assert type(out2_data) != boolalg.Not
        #         out_inv_net = out1
        #         out_net = out2
        #         out_inv_data = out1_data
        #         out_data = out2_data
        #     else:
        #         assert type(out2_data) == boolalg.Not
        #         out_inv_net = out2
        #         out_net = out1
        #         out_inv_data = out2_data
        #         out_data = out1_data
        #
        #     logger.info(f"Non-inverted output: {out_net} = {out_data}")
        #     logger.info(f"Inverted output: {out_inv_net} = {out_inv_data}")
        #
        #     dff.data_in = out_data
        #     dff.data_out = out_net
        #     dff.data_out_inv = out_inv_net
        # else:
        #     assert False

        # Analyze boolean formula of the next flip-flop state.
        # In the simplest case of a D-flip-flop this is just one variable.
        # But there might also be a scan-chain multiplexer or synchronous
        # clear/preset logic.
        data_variables = ff_data_next.atoms(sympy.Symbol)
        num_variables = len(data_variables)

        if num_variables == 1:
            # Simple. Only a data variable, no clear/preset/scan.
            pass
        elif num_variables == 2:
            # There might be a synchronous clear/preset.
            logger.debug(f"Try to detect a synchronous clear or preset from the two data signals {data_variables}.")
            # It is not possible to distinguish data input from clear/preset based only on
            # the boolean formula.
            d, clear, preset = sympy.symbols('d clear preset')
            with_clear = d & clear
            with_preset = d | preset
            if find_boolean_isomorphism(ff_data_next, with_clear) is not None:
                logger.info("Detected synchronous clear.")
            if find_boolean_isomorphism(ff_data_next, with_preset) is not None:
                logger.info("Detected synchronous preset.")
        elif num_variables == 3:
            # Either synchronous clear/preset or scan.
            logger.debug("Try to detect scan-mux.")
            d, scan_enable, scan_in = sympy.symbols('d scan_enable scan_in')

            scan_mux = (scan_in & scan_enable) | (d & ~scan_enable)

            mapping = find_boolean_isomorphism(scan_mux, ff_data_next)
            if mapping is not None:
                logger.info(f"Detected scan-chain mux: {mapping}")
                dff.scan_in = mapping[scan_in]
                dff.scan_enable = mapping[scan_enable]

        else:
            logger.warning(f"Flip-flop data depends on {num_variables} variables."
                           f" Cannot distinguish scan-mux, clear, preset.")

        return dff


def extract_sequential_circuit(c: AbstractCircuit) -> Optional[Union[Latch, SingleEdgeDFF]]:
    logger.debug("Recognize sequential cells.")
    logger.debug(f"Combinational formulas: {c.outputs}")
    logger.debug(f"Latches: {c.latches}")

    extractors = [LatchExtractor(), DFFExtractor()]

    for ex in extractors:
        result = ex.extract(c)
        if result is not None:
            return result
    return None



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Simple sub-process based NgSpice binding.
"""

import os
import subprocess
import queue
import threading
import logging
import numpy as np
from typing import Dict, List, Tuple, Union, Optional
from .piece_wise_linear import PieceWiseLinear

logger = logging.getLogger(__name__)


def run_simulation(sim_file: str, ngspice_executable: str = 'ngspice'):
    """
    Invoke 'ngspice' to run the `sim_file`.
    :param sim_file: Path to ngspice simulation file.
    :return: Returns (stdout, stderr) outputs of ngspice.
    """
    logger.debug(f"Run simulation: {sim_file}")
    try:
        ret = subprocess.run([ngspice_executable, sim_file], capture_output=True)
        # proc = subprocess.Popen([ngspice_executable, sim_file])
        # logger.debug(f"Subprocess return value: {ret}")
        if ret.returncode != 0:
            ngspice_err_message = ret.stderr.decode("utf-8")
            logger.error(f"ngspice simulation failed: {ngspice_err_message}")
            raise Exception(f"ngspice simulation failed: {ngspice_err_message}")

        return ret.stdout.decode("utf-8"), ret.stderr.decode("utf-8")
    except FileNotFoundError as e:
        msg = f"SPICE simulator executable not found. Make sure it is in the current path: {ngspice_executable}"
        logger.error(msg)
        raise FileNotFoundError(msg)


def simulate_cell(
        cell_name: str,
        cell_ports: List[str],
        input_voltages: Dict[str, Union[PieceWiseLinear, float]],
        initial_voltages: Dict[str, float],
        breakpoint_statements: List[str],
        output_voltages: List[str],
        output_currents: List[str],
        simulation_file: str,
        simulation_output_file: str,
        max_simulation_time: float,
        simulation_title: str = "<UNTITLED SIMULATION>",
        temperature: float = 25,
        output_load_capacitances: Dict[str, float] = None,
        time_step: float = 100.0e-12,
        setup_statements: List[str] = None,
        ground_net: str = 'GND',
        debug: bool = False,
) -> Tuple[np.ndarray, Dict[str, np.ndarray], Dict[str, np.ndarray]]:
    """Simulate a circuit with given input signals and measure voltages and currents.

    :param cell_name: Name of the cell to be characterized. Must match with the name used in netlist and liberty.
    :param cell_ports: All circuit pins/ports in the same ordering as used in the SPICE circuit model.
    :param input_voltages: Dictionary holding input voltages or waveforms per net.
    :param initial_voltages: Additional initial voltages per net name at the beginning of the simulation.
        By default initial voltages are derived from the input voltages. This is useful for specifying initial voltages of outputs.
    :param breakpoint_statements: List of ngspice breakpoint statements.
    :param output_voltages: List of voltages that should be in the output of this function.
    :param output_currents: List of sources whose current should be in the output of this function.
    :param simulation_file: Path to the SPICE file which will be generated.
    :param simulation_output_file: Path of the simulation output data.
    :param max_simulation_time: Maximal simulation time in seconds.
    :param simulation_title: Title of the simulation that will be written into the simulation file.
    :param temperature: Temperature of the simulation.
    :param output_load_capacitances: A dict with (net, capacitance) pairs which defines the load capacitances attached to certain nets.
    :param time_step: Simulation time step.
    :param setup_statements: SPICE statements that are included at the beginning of the simulation.
        This should be used for .INCLUDE and .LIB statements.
    :param ground_net: The name of the ground net.
    :param debug: Enable more verbose debugging output such as plots of the simulations.
    :return: Returns tuple (time, Dict[net, voltage], Dict[source, current]).
    The output voltages and currents correspond to the values specified in `output_voltages` and `output_currents`.
    """

    logger.debug(f"Simulate cell: {cell_name}")
    logger.debug(f"Cell ports: {cell_ports}")
    logger.debug(f"Initial voltages: {initial_voltages} [V]")
    logger.debug(f"Ground net: {ground_net}")
    logger.debug(f"Temperature: {temperature} C")
    logger.debug(f"Max. simulation time: {max_simulation_time} s")

    # Create a list of include files.
    if setup_statements is None:
        setup_statements = []

    # Load include files.
    setup_statements_string = "\n".join(setup_statements)

    input_voltages_static = dict()
    input_voltages_active = dict()

    # Split input signals into static and active signals
    for net, voltage in input_voltages.items():
        if isinstance(voltage, PieceWiseLinear):
            input_voltages_active[net] = voltage
        else:
            assert isinstance(voltage, float), "Voltage needs to be a float or a piecewise linear function."
            input_voltages_static[net] = voltage

    logger.debug("Static input voltages: {}".format(sorted(list(input_voltages_static.keys()))))
    logger.debug("Dynamic input voltages: {}".format(sorted(list(input_voltages_active.keys()))))

    def create_voltage_source_statement(net: str, voltage: Union[float, PieceWiseLinear]) -> str:
        """
        Create a SPICE statement for a voltage source driving the 'net' with a voltage.
        The voltage can be either a static value or a `PieceWiseLinear` function.
        """
        if isinstance(voltage, PieceWiseLinear):
            return f"V{net} {net} {ground_net} PWL({voltage.to_spice_pwl_string()}) DC=0"
        elif isinstance(voltage, float):
            return f"V{net} {net} {ground_net} {voltage}"
        else:
            assert False, "`voltage` must be either a float or {}".format(PieceWiseLinear)

    # Create SPICE description of the input voltage sources.
    active_signal_source_statements = "\n".join(
        (
            create_voltage_source_statement(net, wave)
            for net, wave in input_voltages_active.items()
        )
    )

    # Static inputs.
    static_signal_source_statements = "\n".join(
        (
            create_voltage_source_statement(net, voltage)
            for net, voltage in input_voltages_static.items()
        )
    )

    # Load capacitance statements.
    if output_load_capacitances is None:
        output_load_capacitances = dict()
    else:
        assert isinstance(output_load_capacitances, dict)
    load_capacitance_statements = "\n".join(
        (
            f"Cload_{net} {net} {ground_net} {load}"
            for net, load in output_load_capacitances.items()
        )
    )

    # Initial node voltages.
    initial_conditions = initial_voltages

    # Calculate initial conditions for all sources.
    for net, voltage in input_voltages.items():
        if isinstance(voltage, PieceWiseLinear):
            initial_conditions[net] = voltage(0)
        else:
            assert isinstance(voltage, float)
            initial_conditions[net] = voltage

    # Format breakpoint statements.
    breakpoint_statements = '\n'.join(breakpoint_statements)

    # Generate output definitions for voltages.
    output_voltage_defs = " ".join((f"v({v})" for v in output_voltages))

    # Can measure currents only through sources.
    # Check that specified output currents match with specified voltage sources.
    for current in output_currents:
        if current not in input_voltages:
            msg = "Can measure currents only through voltage sources. " \
                  f"'{current}' does not correspond to a source."
            logger.error(msg)
            assert False, msg

    # Generate output definitions for currents.
    output_current_defs = " ".join((f"i(v{source})" for source in output_currents))

    # Create ngspice simulation script.
    sim_netlist = f"""* librecell {__name__}
.title {simulation_title}

.option TEMP={temperature}

{setup_statements_string}

Xcircuit_under_test {" ".join(cell_ports)} {cell_name}

* Output load capacitances.
{load_capacitance_statements}

* Static input and supply voltages.
{static_signal_source_statements}

* Active input signals (clock & data_in).
{active_signal_source_statements}

* Initial conditions.
* Also all voltages of DC sources must be here if they are needed to compute the initial conditions.
.ic {" ".join((f"v({net})={v}" for net, v in initial_conditions.items()))}

.control 
*option reltol=1e-5
*option abstol=1e-15

set filetype=ascii
set wr_vecnames

* Breakpoints
{breakpoint_statements}

* Transient simulation, use initial conditions.
tran {time_step} {max_simulation_time} uic
* Write selected signals to the output file.
wrdata {simulation_output_file} {output_voltage_defs} {output_current_defs} 
* Exit ngspice.
exit
.endc

.end
"""

    # Dump simulation script to the file.
    logger.debug(f"Write simulation netlist: {simulation_file}")
    if os.path.exists(simulation_file):
        logger.debug("Simulation file already exists: {}".format(simulation_file))
    open(simulation_file, "w").write(sim_netlist)

    # Start ngspice.
    logger.debug("Run simulation.")
    stdout, stderr = run_simulation(simulation_file)

    # Retrieve data.
    logger.debug("Load simulation output.")
    sim_data = np.loadtxt(simulation_output_file, skiprows=1)

    if sim_data.ndim != 2:
        logger.error("Simulation failed. No data was written to the output file.")
        if debug:
            logger.error(f"ngspice: {stderr}")
        assert False, "Simulation failed. No data was written to the output file."

    # Extract data from the numpy array.
    time = sim_data[:, 0]
    index = 1

    # Put voltages into a dict.
    voltages = dict()
    for v_out in output_voltages:
        voltages[v_out] = sim_data[:, index]
        index = index + 2

    # Put currents into a dict.
    currents = dict()
    for i_out in output_currents:
        currents[i_out] = sim_data[:, index]
        index = index + 2

    return time, voltages, currents


class NgSpiceInteractive:
    """
    Interface to ngspice as a subprocess.
    """

    def __init__(self,
                 ngspice_executable: str = None,
                 logger: logging.Logger = None):
        if logger is None:
            logger = logging.getLogger(__name__)
        self.logger = logger
        self.ngspice_cmd = 'ngspice' if ngspice_executable is None else ngspice_executable
        self.ngspice_args = ['-p']
        "Run ngspice in interactive pipe mode."

        self.proc: subprocess.Popen[str] = None
        "ngspice process handle."

        self._from_stdout = queue.Queue()
        self._from_stderr = queue.Queue()

        self._stdout_reader_thread = None
        self._stderr_reader_thread = None

    def stop(self):
        if self.proc:
            self.proc.kill()
            self.proc = None

    def start(self):
        self.proc = subprocess.Popen(['ngspice', '-p'],
                                     stdin=subprocess.PIPE,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE,
                                     universal_newlines=True)

        def read_stdout():
            while self.proc:
                line = self.proc.stdout.readline()
                if line is None:
                    break
                line = line.strip()
                if not line:  # Skip empty lines.
                    continue
                # print("stdout:", line)
                self._from_stdout.put(line)

        def read_stderr():
            while self.proc:
                line = self.proc.stderr.readline()
                if line is None:
                    break
                line = line.strip()
                if not line:  # Skip empty lines.
                    continue
                # print("stderr:", line)
                # Forward error messages to the logger.
                if 'Error' in line:
                    self.logger.error(line)
                elif 'Warning' in line:
                    self.logger.warning(line)
                self._from_stderr.put(line)

        self._stdout_reader_thread = threading.Thread(target=read_stdout, daemon=True)
        self._stderr_reader_thread = threading.Thread(target=read_stderr, daemon=True)
        self._stdout_reader_thread.start()
        self._stderr_reader_thread.start()

    def readline(self, timeout: Optional[int] = 1) -> str:
        try:
            return self._from_stdout.get(block=True, timeout=timeout)
        except queue.Empty:
            return None

    def drop_stdout(self):
        """
        Delete the stdout buffer.
        """
        while not self._from_stdout.empty():
            self._from_stdout.get_nowait()

    def drop_stderr(self):
        """
        Delete the stderr buffer.
        """
        while not self._from_stdout.empty():
            self._from_stdout.get_nowait()

    def readline_err(self) -> str:
        try:
            return self._from_stderr.get(block=True, timeout=1)
        except queue.Empty:
            return None

    def _write(self, data: str):
        self.proc.stdin.write(data)

    def _flush(self):
        self.proc.stdin.flush()

    def cmd(self, cmd: str):
        self._write(cmd)
        self._write('\n')
        self._flush()

    def source(self, file: str):
        """
        Load a netlist.
        """
        self.cmd(f"source {file}")

    def reset(self):
        self.cmd('reset')

    def tran(self, t_step, t_stop, t_start=None, uic='') -> int:
        """

        :param t_step:
        :param t_stop:
        :param t_start:
        :param uic:
        :return: Return number of data points.
        """
        self._flush()
        self.cmd(f'tran {t_step} {t_stop} {uic}')
        self.cmd(f'print length(time)')  # Should print something like: length(time) = 1.234e+03

        # Find number of data points.
        num_rows = 0
        while True:
            line = self.readline(timeout=None)
            assert line is not None
            if line.startswith('length(time)'):
                _, num_rows_str = line.split('=', 2)
                num_rows = int(float(num_rows_str))
                break

        assert num_rows > 0
        return num_rows

    def load_circuit(self, circuit: str):
        """
        Load a circuit over stdin.
        :param circuit:
        :return:
        """
        lines = circuit.split('\n')
        for line in lines:
            line = line.strip()
            if line:
                self._write("circbyline ")
                self._write(line)
                self._write('\n')
                self._flush()

    def get_data(self, num_rows: int, voltages: List[str] = None, currents: List[str] = None) -> np.ndarray:
        self.drop_stdout()

        signals = []
        if voltages is not None:
            signals.extend((f"v({v})" for v in voltages))

        if currents is not None:
            signals.extend((f"i({i})" for i in currents))

        signals_str = " ".join(signals)

        self.cmd(f"set width={(1 + len(signals)) * 40}")
        self.cmd(f"print {signals_str}")

        rows = []
        i = 0
        while True:
            line = self.readline()
            assert line is not None
            if line.startswith(str(i)):
                i = i + 1
                arr = np.fromstring(line, dtype=float, sep=' ')
                rows.append(arr)
            if i == num_rows:
                break

        data = np.array(rows)
        # Remove index.
        data = data[:, 1:]

        return data


def simulate_cell_v2(
        cell_name: str,
        cell_ports: List[str],
        input_voltages: Dict[str, Union[PieceWiseLinear, float]],
        initial_voltages: Dict[str, float],
        breakpoint_statements: List[str],
        output_voltages: List[str],
        output_currents: List[str],
        simulation_file: str,
        simulation_output_file: str,
        max_simulation_time: float,
        simulation_title: str = "<UNTITLED SIMULATION>",
        temperature: float = 25,
        output_load_capacitances: Dict[str, float] = None,
        time_step: float = 100.0e-12,
        setup_statements: List[str] = None,
        ground_net: str = 'GND',
        debug: bool = False,
) -> Tuple[np.ndarray, Dict[str, np.ndarray], Dict[str, np.ndarray]]:
    """Simulate a circuit with given input signals and measure voltages and currents.

    :param cell_name: Name of the cell to be characterized. Must match with the name used in netlist and liberty.
    :param cell_ports: All circuit pins/ports in the same ordering as used in the SPICE circuit model.
    :param input_voltages: Dictionary holding input voltages or waveforms per net.
    :param initial_voltages: Additional initial voltages per net name at the beginning of the simulation.
        By default initial voltages are derived from the input voltages. This is useful for specifying initial voltages of outputs.
    :param breakpoint_statements: List of ngspice breakpoint statements.
    :param output_voltages: List of voltages that should be in the output of this function.
    :param output_currents: List of sources whose current should be in the output of this function.
    :param simulation_file: Path to the SPICE file which will be generated.
    :param simulation_output_file: Path of the simulation output data.
    :param max_simulation_time: Maximal simulation time in seconds.
    :param simulation_title: Title of the simulation that will be written into the simulation file.
    :param temperature: Temperature of the simulation.
    :param output_load_capacitances: A dict with (net, capacitance) pairs which defines the load capacitances attached to certain nets.
    :param time_step: Simulation time step.
    :param setup_statements: SPICE statements that are included at the beginning of the simulation.
        This should be used for .INCLUDE and .LIB statements.
    :param ground_net: The name of the ground net.
    :param debug: Enable more verbose debugging output such as plots of the simulations.
    :return: Returns tuple (time, Dict[net, voltage], Dict[source, current]).
    The output voltages and currents correspond to the values specified in `output_voltages` and `output_currents`.
    """

    logger.debug(f"Simulate cell: {cell_name}")
    logger.debug(f"Cell ports: {cell_ports}")
    logger.debug(f"Initial voltages: {initial_voltages} [V]")
    logger.debug(f"Ground net: {ground_net}")
    logger.debug(f"Temperature: {temperature} C")
    logger.debug(f"Max. simulation time: {max_simulation_time} s")

    # Create a list of include files.
    if setup_statements is None:
        setup_statements = []

    # Load include files.
    setup_statements_string = "\n".join(setup_statements)

    input_voltages_static = dict()
    input_voltages_active = dict()

    # Split input signals into static and active signals
    for net, voltage in input_voltages.items():
        if isinstance(voltage, PieceWiseLinear):
            input_voltages_active[net] = voltage
        else:
            assert isinstance(voltage, float), "Voltage needs to be a float or a piecewise linear function."
            input_voltages_static[net] = voltage

    logger.debug("Static input voltages: {}".format(sorted(list(input_voltages_static.keys()))))
    logger.debug("Dynamic input voltages: {}".format(sorted(list(input_voltages_active.keys()))))

    def create_voltage_source_statement(net: str, voltage: Union[float, PieceWiseLinear]) -> str:
        """
        Create a SPICE statement for a voltage source driving the 'net' with a voltage.
        The voltage can be either a static value or a `PieceWiseLinear` function.
        """
        if isinstance(voltage, PieceWiseLinear):
            return f"V{net} {net} {ground_net} PWL({voltage.to_spice_pwl_string()}) DC=0"
        elif isinstance(voltage, float):
            return f"V{net} {net} {ground_net} {voltage}"
        else:
            assert False, "`voltage` must be either a float or {}".format(PieceWiseLinear)

    # Create SPICE description of the input voltage sources.
    active_signal_source_statements = "\n".join(
        (
            create_voltage_source_statement(net, wave)
            for net, wave in input_voltages_active.items()
        )
    )

    # Static inputs.
    static_signal_source_statements = "\n".join(
        (
            create_voltage_source_statement(net, voltage)
            for net, voltage in input_voltages_static.items()
        )
    )

    # Load capacitance statements.
    if output_load_capacitances is None:
        output_load_capacitances = dict()
    else:
        assert isinstance(output_load_capacitances, dict)
    load_capacitance_statements = "\n".join(
        (
            f"Cload_{net} {net} {ground_net} {load}"
            for net, load in output_load_capacitances.items()
        )
    )

    # Initial node voltages.
    initial_conditions = initial_voltages

    # Calculate initial conditions for all sources.
    for net, voltage in input_voltages.items():
        if isinstance(voltage, PieceWiseLinear):
            initial_conditions[net] = voltage(0)
        else:
            assert isinstance(voltage, float)
            initial_conditions[net] = voltage

    # Format breakpoint statements.
    breakpoint_statements = '\n'.join(breakpoint_statements)

    # Generate output definitions for voltages.
    output_voltage_defs = " ".join((f"v({v})" for v in output_voltages))

    # Can measure currents only through sources.
    # Check that specified output currents match with specified voltage sources.
    for current in output_currents:
        if current not in input_voltages:
            msg = "Can measure currents only through voltage sources. " \
                  f"'{current}' does not correspond to a source."
            logger.error(msg)
            assert False, msg

    # Generate output definitions for currents.
    output_current_defs = " ".join((f"i(v{source})" for source in output_currents))

    # Create ngspice simulation script.
    sim_netlist = f"""* librecell {__name__}
.title {simulation_title}

.option TEMP={temperature}

{setup_statements_string}

Xcircuit_under_test {" ".join(cell_ports)} {cell_name}

* Output load capacitances.
{load_capacitance_statements}

* Static input and supply voltages.
{static_signal_source_statements}

* Active input signals (clock & data_in).
{active_signal_source_statements}

* Initial conditions.
* Also all voltages of DC sources must be here if they are needed to compute the initial conditions.
.ic {" ".join((f"v({net})={v}" for net, v in initial_conditions.items()))}

.control 
*option reltol=1e-5
*option abstol=1e-15

set filetype=ascii
set wr_vecnames

* Breakpoints
{breakpoint_statements}

* Transient simulation, use initial conditions.
* tran {time_step} {max_simulation_time} uic
* Write selected signals to the output file.
* wrdata {simulation_output_file} {output_voltage_defs} {output_current_defs} 
* Exit ngspice.
*exit
.endc

.end
"""

    # Dump simulation script to the file.
    if debug:
        logger.debug(f"Write simulation netlist: {simulation_file}")
        if os.path.exists(simulation_file):
            logger.debug("Simulation file already exists: {}".format(simulation_file))
        open(simulation_file, "w").write(sim_netlist)

    # Start ngspice.
    logger.debug("Run simulation.")
    # stdout, stderr = run_simulation(simulation_file)
    ns = NgSpiceInteractive()
    ns.start()

    # ns.source(simulation_file)
    ns.load_circuit(sim_netlist)
    nrows = ns.tran(t_step=time_step, t_stop=max_simulation_time, uic='uic')
    measure_currents_at = [f'v{c}' for c in output_currents]
    sim_data = ns.get_data(num_rows=nrows, voltages=output_voltages, currents=measure_currents_at)
    ns.cmd('quit')
    ns.stop()

    # # Retrieve data.
    # logger.debug("Load simulation output.")
    # sim_data = np.loadtxt(simulation_output_file, skiprows=1)

    if sim_data.ndim != 2:
        logger.error("Simulation failed. No data was written to the output file.")
        if debug:
            logger.error(f"ngspice: {stderr}")
        assert False, "Simulation failed. No data was written to the output file."

    # Extract data from the numpy array.

    assert sim_data.shape[1] == 1 + len(output_voltages) + len(output_currents)
    time = sim_data[:, 0]
    index = 1

    # Put voltages into a dict.
    voltages = dict()
    for v_out in output_voltages:
        voltages[v_out] = sim_data[:, index]
        index = index + 1

    # Put currents into a dict.
    currents = dict()
    for i_out in output_currents:
        currents[i_out] = sim_data[:, index]
        index = index + 1

    return time, voltages, currents



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Classes for handling piece-wise linear waveforms.
"""

import numpy as np
from scipy import interpolate
from typing import Tuple, Sequence, Iterable, Union
from itertools import chain
import itertools


class PieceWiseLinear:

    def __init__(self, x: Iterable, y: Iterable):
        """
        Define a pice wise linear wave by x,y coordinates.
        :param x: x (time) coordinate
        :param y: y (amplitude) coordinate
        """

        assert all(a <= b for a, b in zip(x, x[1:])), "x axis must be sorted."

        self.x = np.array(x)
        self.y = np.array(y)

    def concatenate(self, other):
        # TODO: untested
        return PieceWiseLinear(
            np.concatenate([self.x, other.x + self.x[-1]]),
            np.concatenate([self.y, other.y])
        )

    def split_at(self, x: Union[int, float]) -> Tuple:
        # TODO: untested
        index = np.max(np.argwhere(self.x < x) + 1, initial=0)
        self.add_sampling_point(x)

        return PieceWiseLinear(self.x[:index + 1], self.y[:index + 1]), \
               PieceWiseLinear(self.x[index:], self.y[:index])

    def interpolated(self):
        """
        Represent the wave form as a function of x.
        Interpolation: Linear
        Extrapolation: Using the closest defined value.
        :return: Function of x
        """
        return interpolate.interp1d(self.x, self.y,
                                    bounds_error=False,
                                    fill_value=(self.y[0], self.y[-1]))

    def add_sampling_point(self, x: Union[int, float]):
        """
        Inserts a redundant (x,y) sample point into the wave. y is determined by interpolation
        or extrapolation. The waveform itself is not changed.
        This is used if the raw x,y arrays are required to be defined at certain intervals.
        :param x: x offset
        :return:
        """

        if x in self.x:
            return

        y = self(x)
        index = np.max(np.argwhere(self.x < x) + 1, initial=0)

        self.x = np.insert(self.x, index, x)
        self.y = np.insert(self.y, index, y)

    def delayed(self, delta_t):
        """
        Get the same wave shifted by `delta_t`.
        :param delta_t: Time shift.
        :return: f(t + delta_t)
        """
        return PieceWiseLinear(
            x=self.x + delta_t,
            y=self.y.copy()
        )

    def __add__(self, other):
        """
        Add two functions.
        Input functions are assumed to be 0 ouside of their definition range.
        :param other:
        :return: PieceWiseLinearWave
        """

        if isinstance(other, PieceWiseLinear):

            f1 = self.interpolated()
            f2 = other.interpolated()

            # Concatenate, remove duplicates and sort by time.
            x = set(chain(self.x, other.x))
            x = np.array(sorted(set(x)))

            y = f1(x) + f2(x)

            return PieceWiseLinear(x, y)
        else:
            return PieceWiseLinear(self.x.copy(), self.y + other)

    def __radd__(self, other):
        return self + other

    def __mul__(self, other):
        """
        Multiplication by a scalar.
        :param other: Scalar
        :return: PieceWiseLinearWave
        """
        assert isinstance(other, float) or isinstance(other, int)
        return PieceWiseLinear(self.x.copy(), self.y * other)

    def __rmul__(self, other):
        return self * other

    def __neg__(self):
        negy = -self.y
        return PieceWiseLinear(self.x.copy(), negy)

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return -self + other

    def __call__(self, x):
        return self.interpolated()(x)

    def to_spice_pwl_string(self):
        """
        Format the signal as it is needed for describing a PWL source in SPICE.
        The format is like: "T1 V1 T2 V2 ..." Where T is a time in seconds, V a voltage in volts.
        :return:
        """
        pwl_string = ' '.join((
            f'{time:0.20e}s {voltage:0.20e}V'
            for time, voltage in zip(self.x, self.y)
        ))
        return pwl_string


class StepWave(PieceWiseLinear):
    def __init__(self, start_time,
                 polarity: bool = True,
                 transition_time=0.0,
                 rise_threshold=0.5,
                 fall_threshold=0.5):
        assert 0 <= rise_threshold <= 1
        assert 0 <= fall_threshold <= 1
        assert transition_time >= 0

        transition_threshold = rise_threshold if polarity else 1 - fall_threshold

        start = start_time - transition_threshold * transition_time
        end = start_time + ((1 - transition_threshold) * transition_time)

        if polarity:
            y = [0, 1]
        else:
            y = [1, 0]

        super().__init__(
            x=[start, end],
            y=y
        )


def test_step_wave():
    from math import isclose
    rising_edge = StepWave(start_time=0, polarity=True, transition_time=1,
                           rise_threshold=0.8, fall_threshold=0.2)

    assert isclose(rising_edge(-0.8), 0)
    assert rising_edge(-0.79999) > 0
    assert isclose(rising_edge(0), 0.8)
    assert rising_edge(0.19999) < 1
    assert isclose(rising_edge(0.2), 1)

    falling_edge = StepWave(start_time=0, polarity=False, transition_time=1,
                            rise_threshold=0.8, fall_threshold=0.2)

    assert isclose(falling_edge(-0.8), 1)
    assert falling_edge(-0.79999) < 1
    assert isclose(falling_edge(0), 0.2)
    assert falling_edge(0.19999) > 0
    assert isclose(falling_edge(0.2), 0)


def test_step_wave_addition():
    # Add two step waves where the transitions are overlapping
    from math import isclose
    rising_edge = StepWave(start_time=0, polarity=True, transition_time=1,
                           rise_threshold=0.8, fall_threshold=0.2)

    falling_edge = StepWave(start_time=0.5, polarity=False, transition_time=1,
                            rise_threshold=0.8, fall_threshold=0.2)

    pulse = rising_edge + (falling_edge - 1)

    # import matplotlib.pyplot as plt
    # x = np.linspace(-1, 2, 1000)
    # y = pulse(x)
    # plt.plot(x, y)
    # plt.show()

    assert isclose(pulse(0), 0.5)
    assert isclose(pulse(1), 0)


class PulseWave(PieceWiseLinear):

    # TODO: also use slew thresholds. Maybe pass TripPoint object to constructor.
    def __init__(self, start_time: float,
                 duration: float,
                 polarity: bool = True,
                 rise_time=0.0,
                 fall_time=0.0,
                 rise_threshold=0.5,
                 fall_threshold=0.5):
        """
        Create a single pulse with given start time and duration.
        The duration is measured from the crossing of rise_threshold to fall_threshold.

        :param start_time:
        :param duration:
        :param polarity: True: create a HIGH pulse, False: create a LOW pulse.
        :param rise_time:
        :param fall_time:
        :param rise_threshold:
        :param fall_threshold:
        """
        assert 0 <= rise_threshold <= 1
        assert 0 <= fall_threshold <= 1
        assert duration >= 0

        if polarity:
            rising_edge = StepWave(start_time=start_time, polarity=True, transition_time=rise_time,
                                   rise_threshold=rise_threshold, fall_threshold=fall_threshold)

            falling_edge = StepWave(start_time=start_time + duration, polarity=False, transition_time=fall_time,
                                    rise_threshold=rise_threshold, fall_threshold=fall_threshold)

            pulse = rising_edge + (falling_edge - 1)
            # start1 = start_time - rise_threshold * rise_time
            # end1 = start_time + ((1 - rise_threshold) * rise_time)
            # start2 = start_time + duration - (1 - fall_threshold) * fall_time
            # end2 = start_time + duration + fall_threshold * fall_time
        else:

            falling_edge = StepWave(start_time=start_time, polarity=False, transition_time=fall_time,
                                    rise_threshold=rise_threshold, fall_threshold=fall_threshold)

            rising_edge = StepWave(start_time=start_time + duration, polarity=True, transition_time=rise_time,
                                   rise_threshold=rise_threshold, fall_threshold=fall_threshold)

            pulse = falling_edge + rising_edge

            # start1 = start_time - (1 - fall_threshold) * fall_time
            # end1 = start_time + fall_threshold * fall_time
            # start2 = start_time + duration - rise_threshold * rise_time
            # end2 = start_time + duration + (1 - rise_threshold) * rise_time

        # assert start1 <= end1
        # assert start2 <= end2
        #
        # assert start1 <= end1 <= start2 <= end2

        # if polarity:
        #     y = [0, 1, 1, 0]
        # else:
        #     y = [1, 0, 0, 1]

        # super().__init__(
        #     x=[start1, end1, start2, end2],
        #     y=y
        # )

        super().__init__(x=pulse.x, y=pulse.y)


def test_piece_wise_linear():
    p1 = PieceWiseLinear([0, 1, 2, 3],
                         [0, 1, 1, 0])

    p2 = PieceWiseLinear([0, 1, 2, 3, 4],
                         [0, 1, 1, 1, 1])

    p3 = p1 + p2

    p4 = p2 - p1

    p5 = p1 * 1.23

    for x in [0, 0.5, 1, 2, 2.5, 3, 4, 5]:
        assert p3(x) == p1(x) + p2(x)
        assert p4(x) == p2(x) - p1(x)
        assert p5(x) == p1(x) * 1.23


def test_pulse_wave():
    pulse = PulseWave(start_time=10, duration=10, rise_time=2, fall_time=2)
    pulse_inv = PulseWave(start_time=10, duration=10, rise_time=2, fall_time=2, polarity=False)

    f = pulse.interpolated()

    x = [-1, 9, 10, 11, 15, 19, 20, 21, 100]
    expected = np.array([0, 0, 0.5, 1, 1, 1, 0.5, 0, 0])
    expected_inv = 1 - expected

    actual = f(x)

    assert all(expected == actual)
    assert all(expected_inv == pulse_inv.interpolated()(x))


def test_pulse_wave_negative():
    from math import isclose
    p = PulseWave(start_time=0,
                  duration=10,
                  rise_time=1,
                  fall_time=2,
                  polarity=False,
                  rise_threshold=0.8,
                  fall_threshold=0.2)

    assert isclose(p(-2), 1)
    assert isclose(p(-1.6), 1)
    assert p(-0.7999) < 1
    assert isclose(p(0), 0.2)
    assert isclose(p(5), 0)
    assert isclose(p(10), 0.8)
    assert p(10.1999) < 1
    assert isclose(p(10.2), 1)


def test_pulse_wave_positive():
    from math import isclose
    p = PulseWave(start_time=0,
                  duration=10,
                  rise_time=2,
                  fall_time=1,
                  polarity=True,
                  rise_threshold=0.8,
                  fall_threshold=0.2)

    assert isclose(p(-2), 0)
    assert isclose(p(-1.6), 0)
    assert p(-0.7999) > 0
    assert isclose(p(0), 0.8)
    assert isclose(p(5), 1)
    assert isclose(p(10), 0.2)
    assert p(10.1999) > 0
    assert isclose(p(10.2), 0)


def bitsequence_to_piece_wise_linear(bits: Sequence[int],
                                     bit_duration: float,
                                     rise_time: float = 0,
                                     fall_time: float = 0,
                                     rise_threshold: float = 0.0,
                                     fall_threshold: float = 1.0,
                                     start_time: float = 0) -> PieceWiseLinear:
    """ Generate a piece wise linear waveform from a sequence of bits.

    :param bits: Sequence of bits {0,1}*.
    :param bit_duration: Duration of one bit.
    :param rise_time: Required time to change from 0 to 1.
    :param fall_time: Required time to change from 1 to 0.
    :return: A tuple of two `np.ndarray`s (time, voltage).
    """
    pulse_durations = [(bit, len(list(g))) for bit, g in itertools.groupby(bits)]

    durations = []
    start_times = []
    t = 0
    for bit, l in pulse_durations:
        if bit:
            assert l > 0
            durations.append(l * bit_duration)
            start_times.append(t * bit_duration + start_time)
        t += l

    pulses = [PulseWave(start_time=start,
                        duration=duration,
                        rise_time=rise_time,
                        fall_time=fall_time,
                        rise_threshold=rise_threshold,
                        fall_threshold=fall_threshold)
              for start, duration in zip(start_times, durations)]

    wave = sum(pulses)
    wave.add_sampling_point(0)
    # wave.add_sampling_point(10)
    return wave


def bitsequence_to_piece_wise_linear_old(bits: Sequence[int],
                                         bit_duration: float,
                                         rise_time: float = 0,
                                         fall_time: float = 0,
                                         start_time: float = 0) -> PieceWiseLinear:
    """ Generate a piece wise linear waveform from a sequence of bits.

    :param bits: Sequence of bits {0,1}*.
    :param bit_duration: Duration of one bit.
    :param rise_time: Required time to change from 0 to 1.
    :param fall_time: Required time to change from 1 to 0.
    :return: A tuple of two `np.ndarray`s (time, voltage).
    """
    previous = 0
    times = [min(start_time, 0)]  # Make sure voltage is always defined starting from time 0.
    voltages = [previous]
    for i, b in enumerate(bits):
        if b != previous:
            t0 = i * bit_duration + start_time

            times.append(t0)
            voltages.append(previous)

            t1 = t0 + (rise_time if b else fall_time)
            times.append(t1)
            voltages.append(b)

            previous = b

    # # Hold the signal at the last value for a second.
    # times.append(start_time + len(bits) * bit_duration + 1)
    # voltages.append(previous)

    return PieceWiseLinear(np.array(times), np.array(voltages))


def test_bitsequence_to_piece_wise_linear():
    bits = [0, 1, 1, 0]
    wave = bitsequence_to_piece_wise_linear(bits, 10, rise_time=1, fall_time=2,
                                            rise_threshold=0, fall_threshold=1)

    assert all(wave.x == np.array([0, 10, 11, 30, 32]))
    assert all(wave.y == np.array([0, 0, 1, 1, 0]))



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

# # OBSOLETE: Interesting approach, but not up to date with the rest of the program.
# from scipy import optimize
#
# from PySpice.Spice.Netlist import Circuit, SubCircuit
#
# from PySpice.Spice.BasicElement import Mosfet
#
# import PySpice.Logging.Logging as Logging
#
# from ..liberty.util import get_pin_information
# from ..characterization.util import read_trip_points_from_liberty
# from ..lccommon.net_util import load_subcircuit
# from ..characterization.timing_combinatorial import measure_comb_cell
# from liberty.parser import parse_liberty
# from liberty.types import select_cell
# import argparse
# from typing import List
#
# import numpy as np
# import os
# import logging
#
# pyspice_logger = Logging.setup_logging()
#
# logger = logging.getLogger(__name__)
#
#
# def main():
#     """
#     Command-line tool for automated transistor sizing.
#     Currently only combinatorial cells are supported excluding tri-state cells.
#     :return:
#     """
#
#     parser = argparse.ArgumentParser(
#         description='Optimize transistor sizes of a combinatorial cell.'
#                     'Currently more a proof-of-concept than a useful tool.',
#         epilog='')
#
#     parser.add_argument('-l', '--liberty', required=True, metavar='LIBERTY', type=str,
#                         help='Liberty file. Used for definition of trip points.')
#
#     parser.add_argument('--cell', required=True, metavar='CELL_NAME', type=str, help='Cell name.')
#
#     parser.add_argument('--spice', required=True, metavar='SPICE', type=str,
#                         help='SPICE netlist containing a subcircuit with the same name as the cell.')
#
#     parser.add_argument('-I', '--include', required=False, action='append', metavar='SPICE_INCLUDE', type=str,
#                         help='SPICE files to include such as transistor models.')
#
#     parser.add_argument('-o', '--output', required=True, metavar='SPICE_OUT', type=str,
#                         help='Output netlist with sized transistors.')
#
#     parser.add_argument('--debug', action='store_true', help='Enable debug mode.')
#
#     # Parse arguments
#     args = parser.parse_args()
#
#     DEBUG = args.debug
#     log_level = logging.DEBUG if DEBUG else logging.INFO
#
#     logging.basicConfig(format='%(module)16s %(levelname)8s: %(message)s', level=log_level)
#
#     cell_name = args.cell
#     lib_file = args.liberty
#     spice_file = args.spice
#     spice_output_file = args.output
#
#     # TODO: find ground and supply net names from spice netlist or liberty.
#     ground = 'gnd'
#     supply = 'vdd'
#
#     # Load subcircuit of cell.
#     logger.info("Load subcircuit: {}".format(spice_file))
#     subcircuit_raw = load_subcircuit(spice_file, cell_name)
#     subcircuit = subcircuit_raw.build()
#
#     logger.info("Reading liberty: {}".format(lib_file))
#     with open(lib_file) as f:
#         data = f.read()
#     library = parse_liberty(data)
#
#     cell_group = select_cell(library, cell_name)
#     assert cell_group.args[0] == cell_name
#     logger.info("Cell: {}".format(cell_name))
#
#     # Get information on pins
#     input_pins, output_pins, output_functions = get_pin_information(cell_group)
#
#     # TODO: let user choose parameters.
#     # Fixed channel length
#     channel_length = 5e-08
#     # Minimal and maximal channel width.
#     min_transistor_width = 0.1e-6
#     max_transistor_width = 1e-6
#
#     # Load operation voltage and temperature.
#     # TODO: load voltage/temperature from operating_conditions group
#     supply_voltage = library['nom_voltage']
#     temperature = library['nom_temperature']
#     logger.info('Supply voltage = {:f} V'.format(supply_voltage))
#     logger.info('Temperature = {:f} V'.format(temperature))
#
#     # Read trip points from liberty file.
#     trip_points = read_trip_points_from_liberty(library)
#
#     circuit = Circuit('Test circuit', ground=ground)
#
#     spice_include_files = args.include
#
#     # Load include files.
#     for inc in spice_include_files:
#         logger.info("Include '{}'".format(inc))
#
#         if os.path.isfile(inc):
#             circuit.include(inc)
#         else:
#             msg = "No such file: {}".format(inc)
#             logger.error(msg)
#             assert False, msg
#
#     # Power supply.
#     circuit.V('power_vdd', supply, circuit.gnd, supply_voltage @ u_V)
#
#     # Instantiate mosfets
#     model_map = {
#         'nmos': 'NMOS_VTL',
#         'pmos': 'PMOS_VTL'
#     }
#
#     def create_subcircuit(mosfet_sizes: List[float]) -> SubCircuit:
#         """
#         Create a clone of the original subcircuit with adapted transistor sizes.
#         :param mosfet_sizes:
#         :return:
#         """
#         sc = SubCircuit(subcircuit.name, *subcircuit.nodes)
#         mosfets = (e for e in subcircuit.elements if isinstance(e, Mosfet))
#         for i, (t, size) in enumerate(zip(mosfets, mosfet_sizes)):
#             right, gate, left, body = t.nodes
#
#             model = model_map[t.model]
#
#             # TODO: floating body for SOI
#             sc.M(i + 1, right, gate, left, body,
#                  model=model,
#                  width=size,
#                  length=channel_length
#                  )
#
#         return sc
#
#     # mosfets = []
#     # for i, t in enumerate((e for e in subcircuit.elements if isinstance(e, Mosfet))):
#     #     right, gate, left, body = t.nodes
#     #
#     #     model = model_map[t.model]
#     #
#     #     # TODO: floating body for SOI
#     #     m = circuit.M(i + 1, right, gate, left, body,
#     #                   model=model,
#     #                   width=t.width,
#     #                   length=channel_length
#     #                   )
#     #     mosfets.append(m)
#
#     time_resolution = 10 @ u_ps
#
#     def f(widths: List[float]) -> float:
#         """
#         Objective function to me minimized.
#
#         Creates a version of the circuit with the given transistor sizing, simulates it and
#         returns a cost metric.
#         :param widths: Transistor widths. Ordering of widths must correspond to the ordering
#             of MOSFETs in the original subcircuit.
#         :return: Cost of the transistor sizing.
#         """
#
#         # TODO: let user define load capacitance and input slew.
#         input_transition_time = 0.06
#         output_cap = 1.0
#
#         _circuit = circuit.clone()
#         cut = create_subcircuit(widths)
#         _circuit.subcircuit(cut)
#         _circuit.X('CUT', cut.name, *subcircuit.nodes)
#
#         # for mosfet, width in zip(mosfets, widths):
#         #     mosfet.width = width
#
#         print(_circuit)
#
#         objective = 0
#         for output in output_pins:
#             for active_input in input_pins:
#                 r = measure_comb_cell(_circuit,
#                                       inputs_nets=input_pins,
#                                       active_pin=active_input,
#                                       output_net=output,
#                                       output_functions=output_functions,
#                                       input_rise_time=input_transition_time @ u_ns,
#                                       input_fall_time=input_transition_time @ u_ns,
#                                       output_load_capacitance=output_cap @ u_pF,
#                                       vdd=supply_voltage,
#                                       temperature=temperature,
#                                       trip_points=trip_points,
#                                       time_step=time_resolution,
#                                       simulation_duration_hint=1000 @ u_ps)
#
#             diff = ((r['cell_fall'] - r['cell_rise']) * 1e9) ** 2
#             # _objective = diff + r['cell_rise'] * 1e6 + r['cell_fall'] * 1e6
#             #sum2 = (r['cell_fall'] * 1e9) ** 2 + (r['cell_rise'] * 1e9) ** 2
#
#             _objective = diff  # + sum2
#             # objective = diff[0] + r['rise_delay'] * 1e9
#
#             objective += _objective
#
#         # print(objective)
#         return objective
#
#     start_width = np.mean([min_transistor_width, max_transistor_width])
#
#     num_mosfets = len([e for e in subcircuit.elements if isinstance(e, Mosfet)])
#
#     optimal_widths = optimize.minimize(f,
#                                        x0=[start_width] * num_mosfets,
#                                        bounds=[(min_transistor_width, max_transistor_width)] * num_mosfets,
#                                        tol=3e-7)
#
#     if optimal_widths.success:
#
#         logger.info("Optimization done: {}".format(optimal_widths.message))
#         logger.info("Total function evaluations: {}".format(optimal_widths.nfev))
#         final_subcircuit = create_subcircuit(optimal_widths.x)
#
#         with open(spice_output_file, 'w') as f:
#             logger.info("Write SPICE netlist: {}".format(spice_output_file))
#             f.write(str(final_subcircuit))
#     else:
#         logger.error("Optimization failed: {}".format(optimal_widths.message))
#         exit(1)



# Copyright (c) 2019 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from ..logic import functional_abstraction
from ..logic import seq_recognition
from . import util

import argparse
from PySpice.Unit import *
from lccommon import net_util
from ..lccommon.net_util import load_netlist, extract_transistors
import networkx as nx

from ..licence import licence_notice_string_single_line


def main():
    """
    Command-line tool for cell characterization.
    Currently only combinatorial cells are supported excluding tri-state cells.
    :return:
    """

    print(licence_notice_string_single_line())
    print()

    logger = logging.getLogger(__name__)
    logger.info("sp2bool main function")

    parser = argparse.ArgumentParser(
        description='Find boolean formulas that describe the behaviour of a CMOS circuit.',
        epilog='Example: sp2bool --spice INVX1.sp --cell INVX1')

    parser.add_argument('--cell', required=False, nargs="+", metavar='CELL_NAMES', type=str,
                        help='Cell names. If none is specified all cells will be analyzed.')

    parser.add_argument('--spice', required=True, metavar='SPICE', type=str,
                        help='SPICE netlist containing a subcircuit with the same name as the cell.')

    parser.add_argument('--diff', required=False,
                        nargs="+",
                        metavar='DIFFERENTIAL_PATTERN',
                        type=str,
                        help='Specify differential inputs as "NonInverting,Inverting" tuples.'
                             'The placeholder "%%" can be used like "%%_P,%%_N" or "%%,%%_Diff", ...')

    parser.add_argument('--debug', action='store_true', help='Enable debug mode.')
    parser.add_argument('--plot-network', action='store_true',
                        help='Show a plot of the transistor graph for debugging. '
                             'Transistors are edges in the graph. The edges are labelled with the gate net.')

    # Parse arguments
    args = parser.parse_args()

    DEBUG = args.debug
    log_level = logging.DEBUG if DEBUG else logging.INFO

    if DEBUG:
        log_format = '%(module)16s %(levelname)8s: %(message)s'
    else:
        # Also output name of function in DEBUG mode.
        # log_format = '%(module)16s %(funcName)16s %(levelname)8s: %(message)s'
        log_format = '%(message)s'

    logging.basicConfig(format=log_format, level=log_level)

    if args.cell:
        cell_names = set(args.cell)
    else:
        cell_names = None

    # Load netlist of cell
    netlist_path = args.spice
    logger.info('Load netlist: %s', netlist_path)
    netlist = load_netlist(netlist_path)

    if cell_names is None:
        # Process all circuits if no names were given.
        cell_names = [c.name for c in netlist.each_circuit()]

    for cell_name in cell_names:
        circuit = netlist.circuit_by_name(cell_name)
        if circuit is None:
            logger.error(f"No such circuit: {cell_name}")
            exit(1)
        if cell_names is None or cell_name in cell_names:
            logger.info(f"Analyze {cell_name}")
            transistors_abstract, cell_pins = extract_transistors(circuit)
            io_pins = net_util.get_io_pins(cell_pins)

            logger.info(f"Cell pins: {cell_pins}")
            #print(transistors_abstract)

            # Detect power pins.
            # TODO: don't decide based only on net name.
            power_pins = [p for p in cell_pins if net_util.is_power_net(p)]
            assert len(power_pins) == 2, "Expected to have 2 power pins."
            vdd_pins = [p for p in power_pins if net_util.is_supply_net(p)]
            gnd_pins = [p for p in power_pins if net_util.is_ground_net(p)]
            assert len(vdd_pins) == 1, "Expected to find one VDD pin but found: {}".format(vdd_pins)
            assert len(gnd_pins) == 1, "Expected to find one GND pin but found: {}".format(gnd_pins)
            vdd_pin = vdd_pins[0]
            gnd_pin = gnd_pins[0]

            logger.info(f"Supply net: {vdd_pin}")
            logger.info(f"Ground net: {gnd_pin}")

            # Match differential inputs.
            if args.diff is not None:
                differential_inputs = util.find_differential_inputs_by_pattern(args.diff, io_pins)
            else:
                differential_inputs = dict()

            # Sanity check.
            if len(set(differential_inputs.keys())) != len(set(differential_inputs.values())):
                logger.error(f"Mismatch in the mapping of differential inputs.")
                exit(1)

            def _transistors2multigraph(transistors) -> nx.MultiGraph:
                """ Create a graph representing the transistor network.
                    Each edge corresponds to a transistor, each node to a net.
                """
                G = nx.MultiGraph()
                for t in transistors:
                    G.add_edge(t.source_net, t.drain_net, (t.gate_net, t.channel_type))

                connected = list(nx.connected_components(G))
                print(connected)
                logger.debug(f"Number of connected graphs: {len(connected)}")
                assert nx.is_connected(G)
                return G

            # Derive boolean functions for the outputs from the netlist.
            logger.info("Derive boolean functions for the outputs based on the netlist.")
            transistor_graph = _transistors2multigraph(transistors_abstract)
            if args.plot_network:
                import matplotlib.pyplot as plt
                pos = nx.spring_layout(transistor_graph)
                nx.draw(transistor_graph, pos, with_labels=True)
                print(list(transistor_graph.edges(keys=True)))
                edge_labels = {(a, b): gate_net for a, b, (gate_net, channel_type) in transistor_graph.edges(keys=True)}
                nx.draw_networkx_edge_labels(transistor_graph, pos, edge_labels=edge_labels)
                plt.show()
            abstract = functional_abstraction.analyze_circuit_graph(graph=transistor_graph,
                                                                    pins_of_interest=io_pins,
                                                                    constant_input_pins={vdd_pin: True,
                                                                                         gnd_pin: False},
                                                                    differential_inputs=differential_inputs,
                                                                    user_input_nets=None)
            output_functions_deduced = abstract.outputs

            # Convert keys into strings (they are `sympy.Symbol`s now)
            output_functions_deduced = {output.name: comb.function for output, comb in output_functions_deduced.items()}

            # Log deduced output functions.
            for output_name, function in output_functions_deduced.items():
                logger.info("Deduced output function: {} = {}".format(output_name, function))

            # Try to recognize sequential cells.
            seq = seq_recognition.extract_sequential_circuit(abstract)

            if seq is not None:
                print()
                print(seq)
